@@ -21,15 +21,15 @@
 #ifndef LILV_LILV_H
 #define LILV_LILV_H
 
-#include <stdbool.h>
 #include <stddef.h>
 #include <stdint.h>
 #include <stdio.h>
 
 #include "lv2/lv2plug.in/ns/lv2core/lv2.h"
+#include "lv2/lv2plug.in/ns/ext/urid/urid.h"
 
 #ifdef LILV_SHARED
-#    ifdef __WIN32__
+#    ifdef _WIN32
 #        define LILV_LIB_IMPORT __declspec(dllimport)
 #        define LILV_LIB_EXPORT __declspec(dllexport)
 #    else
@@ -47,6 +47,8 @@
 
 #ifdef __cplusplus
 extern "C" {
+#else
+#   include <stdbool.h>
 #endif
 
 #define LILV_NS_DOAP "http://usefulinc.com/ns/doap#"
@@ -74,6 +76,7 @@ typedef struct LilvUIImpl          LilvUI;           /**< Plugin UI. */
 typedef struct LilvNodeImpl        LilvNode;         /**< Typed Value. */
 typedef struct LilvWorldImpl       LilvWorld;        /**< Lilv World. */
 typedef struct LilvInstanceImpl    LilvInstance;     /**< Plugin instance. */
+typedef struct LilvStateImpl       LilvState;        /**< Plugin state. */
 
 typedef void LilvIter;           /**< Collection iterator */
 typedef void LilvPluginClasses;  /**< set<PluginClass>. */
@@ -298,7 +301,8 @@ lilv_node_as_bool(const LilvNode* value);
    <li>LilvUIs (function prefix "lilv_uis_")</li>
    </ul>
 
-   Each collection type supports a similar basic API:
+   Each collection type supports a similar basic API (except LilvPlugins which
+   is internal and thus lacks a free function):
    <ul>
    <li>void PREFIX_free (coll)</li>
    <li>unsigned PREFIX_size (coll)</li>
@@ -582,6 +586,19 @@ void
 lilv_world_load_bundle(LilvWorld* world,
                        LilvNode*  bundle_uri);
 
+/**
+   Load all the data associated with the given @c resource.
+   @param resource Must be a subject (i.e. a URI or a blank node).
+   @return The number of files parsed, or -1 on error
+
+   All accessible data files linked to @c resource with rdfs:seeAlso will be
+   loaded into the world model.
+*/
+LILV_API
+int
+lilv_world_load_resource(LilvWorld*      world,
+                         const LilvNode* resource);
+
 /**
    Get the parent of all other plugin classes, lv2:Plugin.
 */
@@ -785,6 +802,23 @@ LILV_API
 LilvNodes*
 lilv_plugin_get_optional_features(const LilvPlugin* p);
 
+/**
+   Return whether or not a plugin provides a specific extension data.
+*/
+LILV_API
+bool
+lilv_plugin_has_extension_data(const LilvPlugin* p,
+                               const LilvNode*   uri);
+
+/**
+   Get a sequence of all extension data provided by a plugin.
+   This can be used to find which URIs lilv_instance_get_extension_data()
+   will return a value for without instantiating the plugin.
+*/
+LILV_API
+LilvNodes*
+lilv_plugin_get_extension_data(const LilvPlugin* p);
+
 /**
    Get the number of ports on this plugin.
 */
@@ -864,6 +898,31 @@ const LilvPort*
 lilv_plugin_get_port_by_symbol(const LilvPlugin* plugin,
                                const LilvNode*   symbol);
 
+/**
+   Get a port on @c plugin by its lv2:designation.
+
+   The designation of a port describes the meaning, assignment, allocation or
+   role of the port, e.g. "left channel" or "gain".  If found, the port with
+   matching @a port_class and @a designation is be returned, otherwise NULL is
+   returned.  The @a port_class can be used to distinguish the input and output
+   ports for a particular designation.
+*/
+LILV_API
+LilvPort*
+lilv_plugin_get_port_by_designation(const LilvPlugin* plugin,
+                                    const LilvNode*   port_class,
+                                    const LilvNode*   designation);
+
+/**
+   Get the project the plugin is a part of.
+
+   More information about the project can be read via lilv_world_find_nodes(),
+   typically using properties from DOAP (e.g. doap:name).
+*/
+LILV_API
+LilvNode*
+lilv_plugin_get_project(const LilvPlugin* plugin);
+
 /**
    Get the full name of the plugin's author.
    Returns NULL if author name is not present.
@@ -883,7 +942,7 @@ LilvNode*
 lilv_plugin_get_author_email(const LilvPlugin* plugin);
 
 /**
-   Get the email address of the plugin's author.
+   Get the address of the plugin author's home page.
    Returns NULL if author homepage is not present.
    Returned value must be freed by caller.
 */
@@ -902,7 +961,7 @@ bool
 lilv_plugin_is_replaced(const LilvPlugin* plugin);
 
 /**
-   Write the Turtle description of @c plugin to @c file.
+   Write the Turtle description of @c plugin to @c plugin_file.
 
    This function is particularly useful for porting plugins in conjunction with
    an LV2 bridge such as NASPRO.
@@ -914,6 +973,12 @@ lilv_plugin_write_description(LilvWorld*        world,
                               const LilvNode*   base_uri,
                               FILE*             plugin_file);
 
+/**
+   Write a manifest entry for @c plugin to @c manifest_file.
+
+   This function is intended for use with lilv_plugin_write_description to
+   write a complete description of a plugin to a bundle.
+*/
 LILV_API
 void
 lilv_plugin_write_manifest_entry(LilvWorld*        world,
@@ -922,6 +987,23 @@ lilv_plugin_write_manifest_entry(LilvWorld*        world,
                                  FILE*             manifest_file,
                                  const char*       plugin_file_path);
 
+/**
+   Get the resources related to @c plugin with lv2:appliesTo.
+
+   Some plugin-related resources are not linked directly to the plugin with
+   rdfs:seeAlso and thus will not be automatically loaded along with the plugin
+   data (usually for performance reasons).  All such resources of the given @c
+   type related to @c plugin can be accessed with this function.
+
+   If @c type is NULL, all such resources will be returned, regardless of type.
+
+   To actually load the data for each returned resource, use
+   lilv_world_load_resource.
+*/
+LILV_API
+LilvNodes*
+lilv_plugin_get_related(const LilvPlugin* plugin, const LilvNode* type);
+
 /**
    @}
    @name Port
@@ -963,6 +1045,16 @@ lilv_port_supports_event(const LilvPlugin* p,
                          const LilvPort*   port,
                          const LilvNode*   event_uri);
 
+/**
+   Get the index of a port.
+   The index is only valid for the life of the plugin and may change between
+   versions.  For a stable identifier, use the symbol.
+*/
+LILV_API
+uint32_t
+lilv_port_get_index(const LilvPlugin* plugin,
+                    const LilvPort*   port);
+
 /**
    Get the symbol of a port.
    The 'symbol' is a short string, a valid C identifier.
@@ -1038,6 +1130,263 @@ LilvScalePoints*
 lilv_port_get_scale_points(const LilvPlugin* plugin,
                            const LilvPort*   port);
 
+/**
+   @}
+   @name Plugin State
+   @{
+*/
+
+/**
+   Load a state snapshot from @c world's RDF model.
+   @param subject The subject of the state description (e.g. a preset URI).
+   @return A new LilvState which must be freed with lilv_state_free().
+*/
+LILV_API
+LilvState*
+lilv_state_new_from_world(LilvWorld*      world,
+                          LV2_URID_Map*   map,
+                          const LilvNode* subject);
+
+/**
+   Load a state snapshot from a file.
+   @param subject The subject of the state description (e.g. a preset URI).
+   @param path The path of the file containing the state description.
+   @return A new LilvState which must be freed with lilv_state_free().
+
+   If @c subject is NULL, it is taken to be the URI of the file (i.e.
+   "<>" in Turtle).
+
+   This function parses the file separately to create the state, it does not
+   parse the file into the world model, i.e. the returned state is the only
+   new memory consumed once this function returns.
+*/
+LILV_API
+LilvState*
+lilv_state_new_from_file(LilvWorld*      world,
+                         LV2_URID_Map*   map,
+                         const LilvNode* subject,
+                         const char*     path);
+
+/**
+   Load a state snapshot from a string made by lilv_state_to_string().
+*/
+LILV_API
+LilvState*
+lilv_state_new_from_string(LilvWorld*    world,
+                           LV2_URID_Map* map,
+                           const char*   str);
+
+/**
+   Function to get a port value.
+   @param port_symbol The symbol of the port.
+   @param user_data The user_data passed to lilv_state_new_from_instance().
+   @param size (Output) The size of the returned value.
+   @param type (Output) The URID of the type of the returned value.
+   @return A pointer to the port value.
+
+   This function MUST set @p size and @p type appropriately.
+*/
+typedef const void* (*LilvGetPortValueFunc)(const char* port_symbol,
+                                            void*       user_data,
+                                            uint32_t*   size,
+                                            uint32_t*   type);
+
+/**
+   Create a new state snapshot from a plugin instance.
+
+   @param plugin The plugin this state applies to.
+
+   @param instance An instance of @c plugin.
+
+   @param file_dir Directory of files created by the plugin earlier (or NULL).
+   This is for hosts that support file creation at any time with state
+   state:makePath.  These files will be copied as necessary to @c copy_dir and
+   not be referred to directly in state (a temporary directory is appropriate).
+
+   @param copy_dir Directory of copies of files in @c file_dir (or NULL).  This
+   directory will have the same structure as @c file_dir but with possibly
+   modified file names to distinguish different revisions.  If you only care
+   about saving one state snapshot, it can be the same as @c save_dir.  Plugin
+   state will refer to files in this directory.
+
+   @param save_dir Directory of files created by plugin during save (or NULL).
+   If the state will be saved, this should be the bundle directory later passed
+   to lilv_state_save.
+
+   @param link_dir Directory of links to external files (or NULL).  A link will
+   be made in this directory to any external files referred to in plugin state.
+   In turn, links will be created in the save directory to these links (e.g.
+   save_dir/file => link_dir/file => /foo/bar/file).  This allows many state
+   snapshots to share a single link to an external file, so archival
+   (e.g. with tar -h) will not create several copies of the file.  If this is
+   not required, it can be the same as save_dir.
+
+   @param flags Bitwise OR of LV2_State_Flags values.
+
+   @param features Features to pass LV2_State_Interface.save().
+
+   @return A new LilvState which must be freed with lilv_state_free().
+
+   This function may be called simultaneously with any instance function
+   (except discovery functions) unless the threading class of that function
+   explicitly disallows this.
+
+   To support advanced file functionality, there are several directory
+   parameters.  Simple hosts that only wish to save a single plugins state once
+   may simply use the same directory for all of them (or pass NULL to not
+   support files at all).  The multiple parameters are necessary to support
+   saving an instances state many times while avoiding any duplication of data.
+
+   If supported (via state:makePath passed to LV2_Descriptor::instantiate()),
+   @c file_dir should be the directory where any files created by the plugin
+   (not during save time, e.g. during instantiation) are stored.  These files
+   will be copied to preserve their state at this time.plugin-created files are stored.
+   Lilv will assume any files within this directory (recursively) are created
+   by the plugin and all other files are immutable.  Note that this function
+   does not save the state, use lilv_state_save() for that.
+
+   See <a href="http://lv2plug.in/ns/ext/state/state.h">state.h</a> from the
+   LV2 State extension for details on the @c flags and @c features parameters.
+*/
+LILV_API
+LilvState*
+lilv_state_new_from_instance(const LilvPlugin*          plugin,
+                             LilvInstance*              instance,
+                             LV2_URID_Map*              map,
+                             const char*                file_dir,
+                             const char*                copy_dir,
+                             const char*                link_dir,
+                             const char*                save_dir,
+                             LilvGetPortValueFunc       get_value,
+                             void*                      user_data,
+                             uint32_t                   flags,
+                             const LV2_Feature *const * features);
+
+/**
+   Free @c state.
+*/
+LILV_API
+void
+lilv_state_free(LilvState* state);
+
+/**
+   Return true iff @c a is equivalent to @c b.
+*/
+LILV_API
+bool
+lilv_state_equals(const LilvState* a, const LilvState* b);
+
+/**
+   Return the number of properties in @c state.
+*/
+LILV_API
+unsigned
+lilv_state_get_num_properties(const LilvState* state);
+
+/**
+   Get the URI of the plugin @c state applies to.
+*/
+LILV_API
+const LilvNode*
+lilv_state_get_plugin_uri(const LilvState* state);
+
+/**
+   Get the label of @c state.
+*/
+LILV_API
+const char*
+lilv_state_get_label(const LilvState* state);
+
+/**
+   Set the label of @c state.
+*/
+LILV_API
+void
+lilv_state_set_label(LilvState*  state,
+                     const char* label);
+
+/**
+   Function to set a port value.
+   @param port_symbol The symbol of the port.
+   @param user_data The user_data passed to lilv_state_restore().
+   @param size The size of @p value.
+   @param type The URID of the type of @p value.
+   @param value A pointer to the port value.
+*/
+typedef void (*LilvSetPortValueFunc)(const char* port_symbol,
+                                     void*       user_data,
+                                     const void* value,
+                                     uint32_t    size,
+                                     uint32_t    type);
+
+/**
+   Restore a plugin instance from a state snapshot.
+   @param state The state to restore, which must apply to the correct plugin.
+   @param instance An instance of the plugin @c state applies to.
+   @param set_value A function to set a port value (may be NULL).
+   @param flags Bitwise OR of LV2_State_Flags values.
+   @param features Features to pass LV2_State_Interface.restore().
+
+   This will set all the properties of @c instance to the values stored in @c
+   state.  If @c set_value is provided, it will be called (with the given @c
+   user_data) to restore each port value, otherwise the host must restore the
+   port values itself (using lilv_state_get_port_value) in order to completely
+   restore @c state.
+
+   If the state has properties, this function is in the "instantiation"
+   threading class, i.e. it MUST NOT be called simultaneously with any function
+   on the same plugin instance.  If the state has no properties, only port
+   values are set via @c set_value.
+
+   See <a href="http://lv2plug.in/ns/ext/state/state.h">state.h</a> from the
+   LV2 State extension for details on the @c flags and @c features parameters.
+*/
+LILV_API
+void
+lilv_state_restore(const LilvState*           state,
+                   LilvInstance*              instance,
+                   LilvSetPortValueFunc       set_value,
+                   void*                      user_data,
+                   uint32_t                   flags,
+                   const LV2_Feature *const * features);
+
+/**
+   Save state to a file.
+   @param unmap URID unmapper.
+   @param state State to save.
+   @param uri URI of state, may be NULL.
+   @param dir Path of the bundle directory to save into.
+   @param filename Path of the state file relative to @c dir.
+
+   The format of state on disk is compatible with that defined in the LV2
+   preset extension, i.e. this function may be used to save presets which can
+   be loaded by any host.
+
+   If @c uri is NULL, the preset URI will be a file URI, but the bundle
+   can safely be moved (i.e. the state file will use "<>" as the subject).
+*/
+LILV_API
+int
+lilv_state_save(LilvWorld*                 world,
+                LV2_URID_Map*              map,
+                LV2_URID_Unmap*            unmap,
+                const LilvState*           state,
+                const char*                uri,
+                const char*                dir,
+                const char*                filename);
+
+/**
+   Save state to a string.  This function does not use the filesystem.
+*/
+LILV_API
+char*
+lilv_state_to_string(LilvWorld*       world,
+                     LV2_URID_Map*    map,
+                     LV2_URID_Unmap*  unmap,
+                     const LilvState* state,
+                     const char*      uri,
+                     const char*      base_uri);
+
 /**
    @}
    @name Scale Point
