@@ -21,7 +21,10 @@
 
 namespace Lilv {
 
-const char* uri_to_path(const char* uri) { return lilv_uri_to_path(uri); }
+static inline const char*
+uri_to_path(const char* uri) {
+	return lilv_uri_to_path(uri);
+}
 
 #define LILV_WRAP0(RT, prefix, name) \
 	inline RT name() { return lilv_ ## prefix ## _ ## name (me); }
@@ -116,7 +119,8 @@ struct PluginClass {
 
 struct PluginClasses {
 	LILV_WRAP_COLL(PluginClasses, PluginClass, plugin_classes);
-	LILV_WRAP1(const PluginClass, plugin_classes, get_by_uri, const LilvNode*, uri);
+	LILV_WRAP1(const PluginClass, plugin_classes,
+	           get_by_uri, const LilvNode*, uri);
 };
 
 struct ScalePoints {
