@@ -21,16 +21,21 @@
 extern "C" {
 #endif
 
-#include <stdbool.h>
 #include <stddef.h>
 #include <stdint.h>
 #include <stdlib.h>
+#include <float.h>
 
-#ifdef __WIN32__
+#ifdef _WIN32
 #    include <windows.h>
 #    define dlopen(path, flags) LoadLibrary(path)
-#    define dlclose(lib) FreeLibrary(lib)
+#    define dlclose(lib) FreeLibrary((HMODULE)lib)
 #    define dlsym GetProcAddress
+#    ifdef _MSC_VER
+#        define __func__ __FUNCTION__
+#        define INFINITY DBL_MAX + DBL_MAX
+#        define NAN INFINITY - INFINITY
+#    endif
 static inline char* dlerror(void) { return "Unknown error"; }
 #else
 #    include <dlfcn.h>
@@ -41,11 +46,11 @@ static inline char* dlerror(void) { return "Unknown error"; }
 
 #include "zix/tree.h"
 
-#include "lilv-config.h"
+#include "lilv_config.h"
 #include "lilv/lilv.h"
 
 #ifdef LILV_DYN_MANIFEST
-#    include "lv2/lv2plug.in/ns/ext/dyn-manifest/dyn-manifest.h"
+#    include "lv2/lv2plug.in/ns/ext/dynmanifest/dynmanifest.h"
 #endif
 
 /*
@@ -65,7 +70,6 @@ struct LilvPortImpl {
 	LilvNodes* classes;  ///< rdf:type
 };
 
-
 struct LilvSpecImpl {
 	SordNode*            spec;
 	SordNode*            bundle;
@@ -83,12 +87,34 @@ struct LilvHeader {
 	LilvNode*  uri;
 };
 
+#ifdef LILV_DYN_MANIFEST
+typedef struct {
+	LilvNode*               bundle;
+	void*                   lib;
+	LV2_Dyn_Manifest_Handle handle;
+	uint32_t                refs;
+} LilvDynManifest;
+#endif
+
+typedef struct {
+	LilvWorld*                world;
+	LilvNode*                 uri;
+	void*                     lib;
+	LV2_Descriptor_Function   lv2_descriptor;
+#ifdef LILV_NEW_LV2
+	const LV2_Lib_Descriptor* desc;
+#endif
+	uint32_t                  refs;
+} LilvLib;
+
 struct LilvPluginImpl {
 	LilvWorld*             world;
 	LilvNode*              plugin_uri;
 	LilvNode*              bundle_uri;  ///< Bundle plugin was loaded from
 	LilvNode*              binary_uri;  ///< lv2:binary
-	LilvNode*              dynman_uri;  ///< dynamic manifest binary
+#ifdef LILV_DYN_MANIFEST
+	LilvDynManifest*       dynmanifest;
+#endif
 	const LilvPluginClass* plugin_class;
 	LilvNodes*             data_uris;  ///< rdfs::seeAlso
 	LilvPort**             ports;
@@ -104,6 +130,11 @@ struct LilvPluginClassImpl {
 	LilvNode*  label;
 };
 
+struct LilvInstancePimpl {
+	LilvWorld* world;
+	LilvLib*   lib;
+};
+
 typedef struct {
 	bool dyn_manifest;
 	bool filter_language;
@@ -118,34 +149,44 @@ struct LilvWorldImpl {
 	LilvPluginClasses* plugin_classes;
 	LilvSpec*          specs;
 	LilvPlugins*       plugins;
-	SordNode*          dc_replaces_node;
-	SordNode*          dyn_manifest_node;
-	SordNode*          lv2_binary_node;
-	SordNode*          lv2_default_node;
-	SordNode*          lv2_index_node;
-	SordNode*          lv2_maximum_node;
-	SordNode*          lv2_minimum_node;
-	SordNode*          lv2_plugin_node;
-	SordNode*          lv2_port_node;
-	SordNode*          lv2_portproperty_node;
-	SordNode*          lv2_reportslatency_node;
-	SordNode*          lv2_specification_node;
-	SordNode*          lv2_symbol_node;
-	SordNode*          rdf_a_node;
-	SordNode*          rdf_value_node;
-	SordNode*          rdfs_class_node;
-	SordNode*          rdfs_label_node;
-	SordNode*          rdfs_seealso_node;
-	SordNode*          rdfs_subclassof_node;
-	SordNode*          xsd_boolean_node;
-	SordNode*          xsd_decimal_node;
-	SordNode*          xsd_double_node;
-	SordNode*          xsd_integer_node;
-	LilvNode*          doap_name_val;
-	LilvNode*          lv2_name_val;
-	LilvNode*          lv2_optionalFeature_val;
-	LilvNode*          lv2_requiredFeature_val;
-	LilvOptions        opt;
+	LilvNodes*         loaded_files;
+	ZixTree*           libs;
+	struct {
+		SordNode* dc_replaces;
+		SordNode* dman_DynManifest;
+		SordNode* doap_name;
+		SordNode* lv2_Plugin;
+		SordNode* lv2_Specification;
+		SordNode* lv2_appliesTo;
+		SordNode* lv2_binary;
+		SordNode* lv2_default;
+		SordNode* lv2_designation;
+		SordNode* lv2_extensionData;
+		SordNode* lv2_index;
+		SordNode* lv2_maximum;
+		SordNode* lv2_minimum;
+		SordNode* lv2_name;
+		SordNode* lv2_optionalFeature;
+		SordNode* lv2_port;
+		SordNode* lv2_portProperty;
+		SordNode* lv2_reportsLatency;
+		SordNode* lv2_requiredFeature;
+		SordNode* lv2_symbol;
+		SordNode* null_uri;
+		SordNode* pset_value;
+		SordNode* rdf_a;
+		SordNode* rdf_value;
+		SordNode* rdfs_Class;
+		SordNode* rdfs_label;
+		SordNode* rdfs_seeAlso;
+		SordNode* rdfs_subClassOf;
+		SordNode* xsd_base64Binary;
+		SordNode* xsd_boolean;
+		SordNode* xsd_decimal;
+		SordNode* xsd_double;
+		SordNode* xsd_integer;
+	} uris;
+	LilvOptions opt;
 };
 
 typedef enum {
@@ -154,7 +195,8 @@ typedef enum {
 	LILV_VALUE_INT,
 	LILV_VALUE_FLOAT,
 	LILV_VALUE_BOOL,
-	LILV_VALUE_BLANK
+	LILV_VALUE_BLANK,
+	LILV_VALUE_BLOB
 } LilvNodeType;
 
 struct LilvNodeImpl {
@@ -165,6 +207,10 @@ struct LilvNodeImpl {
 		float     float_val;
 		bool      bool_val;
 		SordNode* uri_val;
+		struct {
+			void*  buf;
+			size_t size;
+		} blob_val;
 	} val;
 	LilvNodeType  type;
 };
@@ -175,7 +221,7 @@ struct LilvScalePointImpl {
 };
 
 struct LilvUIImpl {
-	LilvWorld*  world;
+	LilvWorld* world;
 	LilvNode*  uri;
 	LilvNode*  bundle_uri;
 	LilvNode*  binary_uri;
@@ -216,6 +262,15 @@ LilvPluginClass* lilv_plugin_class_new(LilvWorld*      world,
 
 void lilv_plugin_class_free(LilvPluginClass* plugin_class);
 
+LilvLib*
+lilv_lib_open(LilvWorld*               world,
+              const LilvNode*          uri,
+              const char*              bundle_path,
+              const LV2_Feature*const* features);
+
+const LV2_Descriptor* lilv_lib_get_plugin(LilvLib* lib, uint32_t index);
+void                  lilv_lib_close(LilvLib* lib);
+
 LilvNodes*         lilv_nodes_new(void);
 LilvPlugins*       lilv_plugins_new(void);
 LilvScalePoints*   lilv_scale_points_new(void);
@@ -244,6 +299,9 @@ int lilv_header_compare_by_uri(const void* a, const void* b, void* user_data);
 int
 lilv_ptr_cmp(const void* a, const void* b, void* user_data);
 
+int
+lilv_resource_node_cmp(const void* a, const void* b, void* user_data);
+
 struct LilvHeader*
 lilv_collection_get_by_uri(const ZixTree* seq, const LilvNode* uri);
 
@@ -265,40 +323,37 @@ lilv_world_query_values_internal(LilvWorld*      world,
 #define FOREACH_MATCH(iter) \
 	for (; !sord_iter_end(iter); sord_iter_next(iter))
 
-static inline const SordNode*
-lilv_match_subject(SordIter* iter) {
-	SordQuad tup;
-	sord_iter_get(iter, tup);
-	return tup[SORD_SUBJECT];
-}
+LilvNodes* lilv_nodes_from_stream_objects(LilvWorld*    w,
+                                          SordIter*     stream,
+                                          SordQuadIndex field);
 
-static inline const SordNode*
-lilv_match_object(SordIter* iter) {
-	SordQuad tup;
-	sord_iter_get(iter, tup);
-	return tup[SORD_OBJECT];
-}
+char*  lilv_strjoin(const char* first, ...);
+char*  lilv_strdup(const char* str);
+char*  lilv_get_lang(void);
+char*  lilv_expand(const char* path);
+char*  lilv_dirname(const char* path);
+int    lilv_copy_file(const char* src, const char* dst);
+bool   lilv_path_exists(const char* path, void* ignored);
+char*  lilv_path_absolute(const char* path);
+bool   lilv_path_is_absolute(const char* path);
+char*  lilv_get_latest_copy(const char* path, const char* copy_path);
+char*  lilv_path_relative_to(const char* path, const char* base);
+bool   lilv_path_is_child(const char* path, const char* dir);
+int    lilv_flock(FILE* file, bool lock);
+char*  lilv_realpath(const char* path);
+int    lilv_symlink(const char* oldpath, const char* newpath);
+int    lilv_mkdir_p(const char* path);
+char*  lilv_path_join(const char* a, const char* b);
+bool   lilv_file_equals(const char* a_path, const char* b_path);
 
-static inline void
-lilv_match_end(SordIter* iter)
-{
-	sord_iter_free(iter);
-}
+char*
+lilv_find_free_path(const char* in_path,
+                    bool (*exists)(const char*, void*), void* user_data);
 
-static inline bool lilv_matches_next(SordIter* matches) {
-	return sord_iter_next(matches);
-}
-
-static inline bool lilv_matches_end(SordIter* matches) {
-	return sord_iter_end(matches);
-}
-
-LilvNodes* lilv_nodes_from_stream_objects(LilvWorld* w,
-                                          SordIter*  stream);
-
-char* lilv_strjoin(const char* first, ...);
-char* lilv_strdup(const char* str);
-char* lilv_get_lang(void);
+void
+lilv_dir_for_each(const char* path,
+                  void*       data,
+                  void (*f)(const char* path, const char* name, void* data));
 
 typedef void (*VoidFunc)(void);
 
