@@ -14,19 +14,15 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#define _XOPEN_SOURCE 500
-
 #include <assert.h>
-#include <locale.h>
 #include <stdlib.h>
 #include <string.h>
 
 #include "lilv_internal.h"
 
 static void
-lilv_node_set_numerics_from_string(LilvNode* val)
+lilv_node_set_numerics_from_string(LilvNode* val, size_t len)
 {
-	char* locale;
 	char* endptr;
 
 	switch (val->type) {
@@ -35,24 +31,17 @@ lilv_node_set_numerics_from_string(LilvNode* val)
 	case LILV_VALUE_STRING:
 		break;
 	case LILV_VALUE_INT:
-		// FIXME: locale kludge, need a locale independent strtol
-		locale = lilv_strdup(setlocale(LC_NUMERIC, NULL));
-		setlocale(LC_NUMERIC, "POSIX");
 		val->val.int_val = strtol(val->str_val, &endptr, 10);
-		setlocale(LC_NUMERIC, locale);
-		free(locale);
 		break;
 	case LILV_VALUE_FLOAT:
-		// FIXME: locale kludge, need a locale independent strtod
-		locale = lilv_strdup(setlocale(LC_NUMERIC, NULL));
-		setlocale(LC_NUMERIC, "POSIX");
-		val->val.float_val = strtod(val->str_val, &endptr);
-		setlocale(LC_NUMERIC, locale);
-		free(locale);
+		val->val.float_val = serd_strtod(val->str_val, &endptr);
 		break;
 	case LILV_VALUE_BOOL:
 		val->val.bool_val = (!strcmp(val->str_val, "true"));
 		break;
+	case LILV_VALUE_BLOB:
+		val->val.blob_val.buf = serd_base64_decode(
+			(const uint8_t*)val->str_val, len, &val->val.blob_val.size);
 	}
 }
 
@@ -64,7 +53,7 @@ lilv_node_set_numerics_from_string(LilvNode* val)
 LilvNode*
 lilv_node_new(LilvWorld* world, LilvNodeType type, const char* str)
 {
-	LilvNode* val = malloc(sizeof(struct LilvNodeImpl));
+	LilvNode* val = (LilvNode*)malloc(sizeof(LilvNode));
 	val->world = world;
 	val->type  = type;
 
@@ -80,6 +69,7 @@ lilv_node_new(LilvWorld* world, LilvNodeType type, const char* str)
 	case LILV_VALUE_INT:
 	case LILV_VALUE_FLOAT:
 	case LILV_VALUE_BOOL:
+	case LILV_VALUE_BLOB:
 		val->str_val = lilv_strdup(str);
 		break;
 	}
@@ -94,17 +84,18 @@ lilv_node_new_from_node(LilvWorld* world, const SordNode* node)
 	LilvNode*    result       = NULL;
 	SordNode*    datatype_uri = NULL;
 	LilvNodeType type         = LILV_VALUE_STRING;
+	size_t       len          = 0;
 
 	switch (sord_node_get_type(node)) {
 	case SORD_URI:
-		result              = malloc(sizeof(struct LilvNodeImpl));
+		result              = (LilvNode*)malloc(sizeof(LilvNode));
 		result->world       = (LilvWorld*)world;
 		result->type        = LILV_VALUE_URI;
 		result->val.uri_val = sord_node_copy(node);
 		result->str_val     = (char*)sord_node_get_string(result->val.uri_val);
 		break;
 	case SORD_BLANK:
-		result              = malloc(sizeof(struct LilvNodeImpl));
+		result              = (LilvNode*)malloc(sizeof(LilvNode));
 		result->world       = (LilvWorld*)world;
 		result->type        = LILV_VALUE_BLANK;
 		result->val.uri_val = sord_node_copy(node);
@@ -113,26 +104,23 @@ lilv_node_new_from_node(LilvWorld* world, const SordNode* node)
 	case SORD_LITERAL:
 		datatype_uri = sord_node_get_datatype(node);
 		if (datatype_uri) {
-			if (sord_node_equals(datatype_uri, world->xsd_boolean_node))
+			if (sord_node_equals(datatype_uri, world->uris.xsd_boolean))
 				type = LILV_VALUE_BOOL;
-			else if (sord_node_equals(datatype_uri, world->xsd_decimal_node)
-			         || sord_node_equals(datatype_uri, world->xsd_double_node))
+			else if (sord_node_equals(datatype_uri, world->uris.xsd_decimal)
+			         || sord_node_equals(datatype_uri, world->uris.xsd_double))
 				type = LILV_VALUE_FLOAT;
-			else if (sord_node_equals(datatype_uri, world->xsd_integer_node))
+			else if (sord_node_equals(datatype_uri, world->uris.xsd_integer))
 				type = LILV_VALUE_INT;
+			else if (sord_node_equals(datatype_uri,
+			                          world->uris.xsd_base64Binary))
+				type = LILV_VALUE_BLOB;
 			else
 				LILV_ERRORF("Unknown datatype `%s'\n",
 				            sord_node_get_string(datatype_uri));
 		}
-		result = lilv_node_new(world, type, (const char*)sord_node_get_string(node));
-		switch (result->type) {
-		case LILV_VALUE_INT:
-		case LILV_VALUE_FLOAT:
-		case LILV_VALUE_BOOL:
-			lilv_node_set_numerics_from_string(result);
-		default:
-			break;
-		}
+		result = lilv_node_new(
+			world, type, (const char*)sord_node_get_string_counted(node, &len));
+		lilv_node_set_numerics_from_string(result, len);
 		break;
 	default:
 		assert(false);
@@ -181,7 +169,8 @@ LILV_API
 LilvNode*
 lilv_new_bool(LilvWorld* world, bool val)
 {
-	LilvNode* ret = lilv_node_new(world, LILV_VALUE_BOOL, val ? "true" : "false");
+	LilvNode* ret = lilv_node_new(world, LILV_VALUE_BOOL,
+	                              val ? "true" : "false");
 	ret->val.bool_val = val;
 	return ret;
 }
@@ -193,7 +182,7 @@ lilv_node_duplicate(const LilvNode* val)
 	if (val == NULL)
 		return NULL;
 
-	LilvNode* result = malloc(sizeof(struct LilvNodeImpl));
+	LilvNode* result = (LilvNode*)malloc(sizeof(LilvNode));
 	result->world = val->world;
 	result->type = val->type;
 
@@ -251,6 +240,10 @@ lilv_node_equals(const LilvNode* value, const LilvNode* other)
 		return (value->val.float_val == other->val.float_val);
 	case LILV_VALUE_BOOL:
 		return (value->val.bool_val == other->val.bool_val);
+	case LILV_VALUE_BLOB:
+		return (value->val.blob_val.size == other->val.blob_val.size)
+			&& !memcmp(value->val.blob_val.buf, other->val.blob_val.buf,
+			           value->val.blob_val.size);
 	}
 
 	return false; /* shouldn't get here */
@@ -260,48 +253,36 @@ LILV_API
 char*
 lilv_node_get_turtle_token(const LilvNode* value)
 {
-	size_t len    = 0;
-	char*  result = NULL;
-	char*  locale = NULL;
+	size_t   len    = 0;
+	char*    result = NULL;
+	SerdNode node;
 
 	switch (value->type) {
 	case LILV_VALUE_URI:
 		len = strlen(value->str_val) + 3;
-		result = calloc(len, 1);
+		result = (char*)calloc(len, 1);
 		snprintf(result, len, "<%s>", value->str_val);
 		break;
 	case LILV_VALUE_BLANK:
 		len = strlen(value->str_val) + 3;
-		result = calloc(len, 1);
+		result = (char*)calloc(len, 1);
 		snprintf(result, len, "_:%s", value->str_val);
 		break;
 	case LILV_VALUE_STRING:
 	case LILV_VALUE_BOOL:
+	case LILV_VALUE_BLOB:
 		result = lilv_strdup(value->str_val);
 		break;
 	case LILV_VALUE_INT:
-		// INT64_MAX is 9223372036854775807 (19 digits) + 1 for sign
-		// FIXME: locale kludge, need a locale independent snprintf
-		locale = lilv_strdup(setlocale(LC_NUMERIC, NULL));
-		len = 20;
-		result = calloc(len, 1);
-		setlocale(LC_NUMERIC, "POSIX");
-		snprintf(result, len, "%d", value->val.int_val);
-		setlocale(LC_NUMERIC, locale);
+		node   = serd_node_new_integer(value->val.int_val);
+		result = (char*)node.buf;
 		break;
 	case LILV_VALUE_FLOAT:
-		// FIXME: locale kludge, need a locale independent snprintf
-		locale = lilv_strdup(setlocale(LC_NUMERIC, NULL));
-		len = 20;  // FIXME: proper maximum value?
-		result = calloc(len, 1);
-		setlocale(LC_NUMERIC, "POSIX");
-		snprintf(result, len, "%f", value->val.float_val);
-		setlocale(LC_NUMERIC, locale);
+		node = serd_node_new_decimal(value->val.float_val, 8);
+		result = (char*)node.buf;
 		break;
 	}
 
-	free(locale);
-
 	return result;
 }
 
@@ -370,7 +351,7 @@ LILV_API
 const char*
 lilv_node_as_string(const LilvNode* value)
 {
-	return value->str_val;
+	return value ? value->str_val : NULL;
 }
 
 LILV_API
@@ -401,10 +382,11 @@ float
 lilv_node_as_float(const LilvNode* value)
 {
 	assert(lilv_node_is_float(value) || lilv_node_is_int(value));
-	if (lilv_node_is_float(value))
+	if (lilv_node_is_float(value)) {
 		return value->val.float_val;
-	else  // lilv_node_is_int(value)
+	} else {  // lilv_node_is_int(value)
 		return (float)value->val.int_val;
+	}
 }
 
 LILV_API
