@@ -14,8 +14,6 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#define _XOPEN_SOURCE 500
-
 #include <assert.h>
 #include <math.h>
 #include <stdarg.h>
@@ -23,10 +21,11 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include "lilv-config.h"
+#include "lilv_config.h"
 #include "lilv_internal.h"
 
-#define NS_UI   (const uint8_t*)"http://lv2plug.in/ns/extensions/ui#"
+#include "lv2/lv2plug.in/ns/extensions/ui/ui.h"
+
 #define NS_DOAP (const uint8_t*)"http://usefulinc.com/ns/doap#"
 #define NS_FOAF (const uint8_t*)"http://xmlns.com/foaf/0.1/"
 
@@ -35,13 +34,13 @@ LilvPlugin*
 lilv_plugin_new(LilvWorld* world, LilvNode* uri, LilvNode* bundle_uri)
 {
 	assert(bundle_uri);
-	LilvPlugin* plugin = malloc(sizeof(struct LilvPluginImpl));
+	LilvPlugin* plugin = (LilvPlugin*)malloc(sizeof(LilvPlugin));
 	plugin->world        = world;
 	plugin->plugin_uri   = uri;
 	plugin->bundle_uri   = bundle_uri;
 	plugin->binary_uri   = NULL;
 #ifdef LILV_DYN_MANIFEST
-	plugin->dynman_uri   = NULL;
+	plugin->dynmanifest  = NULL;
 #endif
 	plugin->plugin_class = NULL;
 	plugin->data_uris    = lilv_nodes_new();
@@ -53,9 +52,37 @@ lilv_plugin_new(LilvWorld* world, LilvNode* uri, LilvNode* bundle_uri)
 	return plugin;
 }
 
+static void
+lilv_plugin_free_ports(LilvPlugin* p)
+{
+	if (p->ports) {
+		for (uint32_t i = 0; i < p->num_ports; ++i) {
+			lilv_port_free(p, p->ports[i]);
+		}
+		free(p->ports);
+		p->num_ports = 0;
+		p->ports     = NULL;
+	}
+}
+
 void
 lilv_plugin_free(LilvPlugin* p)
 {
+#ifdef LILV_DYN_MANIFEST
+	if (p->dynmanifest && --p->dynmanifest->refs == 0) {
+		typedef int (*CloseFunc)(LV2_Dyn_Manifest_Handle);
+		CloseFunc close_func = (CloseFunc)lilv_dlfunc(p->dynmanifest->lib,
+		                                              "lv2_dyn_manifest_close");
+		if (close_func) {
+			close_func(p->dynmanifest->handle);
+		}
+
+		dlclose(p->dynmanifest->lib);
+		lilv_node_free(p->dynmanifest->bundle);
+		free(p->dynmanifest);
+	}
+#endif
+
 	lilv_node_free(p->plugin_uri);
 	p->plugin_uri = NULL;
 
@@ -65,18 +92,7 @@ lilv_plugin_free(LilvPlugin* p)
 	lilv_node_free(p->binary_uri);
 	p->binary_uri = NULL;
 
-#ifdef LILV_DYN_MANIFEST
-	lilv_node_free(p->dynman_uri);
-	p->dynman_uri = NULL;
-#endif
-
-	if (p->ports) {
-		for (uint32_t i = 0; i < p->num_ports; ++i) {
-			lilv_port_free(p, p->ports[i]);
-		}
-		free(p->ports);
-		p->ports = NULL;
-	}
+	lilv_plugin_free_ports(p);
 
 	lilv_nodes_free(p->data_uris);
 	p->data_uris = NULL;
@@ -92,10 +108,10 @@ lilv_plugin_get_one(const LilvPlugin* p,
 	LilvNode* ret    = NULL;
 	SordIter* stream = lilv_world_query_internal(
 		p->world, subject, predicate, NULL);
-	if (!lilv_matches_end(stream)) {
-		ret = lilv_node_new_from_node(p->world, lilv_match_object(stream));
+	if (!sord_iter_end(stream)) {
+		ret = lilv_node_new_from_node(p->world, sord_iter_get_node(stream, SORD_OBJECT));
 	}
-	lilv_match_end(stream);
+	sord_iter_free(stream);
 	return ret;
 }
 
@@ -138,41 +154,25 @@ lilv_plugin_load(LilvPlugin* p)
 #ifdef LILV_DYN_MANIFEST
 	typedef void* LV2_Dyn_Manifest_Handle;
 	// Load and parse dynamic manifest data, if this is a library
-	if (p->dynman_uri) {
-		const char* lib_path = lilv_uri_to_path(lilv_node_as_string(p->dynman_uri));
-		void* lib = dlopen(lib_path, RTLD_LAZY);
-		if (!lib) {
-			LILV_WARNF("Failed to open dynamic manifest %s\n",
-			           lilv_node_as_string(p->dynman_uri));
-			return;
-		}
-
-		typedef int (*OpenFunc)(LV2_Dyn_Manifest_Handle*, const LV2_Feature *const *);
-		OpenFunc open_func = (OpenFunc)lilv_dlfunc(lib, "lv2_dyn_manifest_open");
-		LV2_Dyn_Manifest_Handle handle = NULL;
-		if (open_func)
-			open_func(&handle, &dman_features);
-
+	if (p->dynmanifest) {
 		typedef int (*GetDataFunc)(LV2_Dyn_Manifest_Handle handle,
 		                           FILE*                   fp,
 		                           const char*             uri);
 		GetDataFunc get_data_func = (GetDataFunc)lilv_dlfunc(
-			lib, "lv2_dyn_manifest_get_data");
+			p->dynmanifest->lib, "lv2_dyn_manifest_get_data");
 		if (get_data_func) {
-			serd_env_set_base_uri(
-				env, sord_node_to_serd_node(p->dynman_uri->val.uri_val));
+			const SordNode* bundle = p->dynmanifest->bundle->val.uri_val;
+			serd_env_set_base_uri(env, sord_node_to_serd_node(bundle));
 			FILE* fd = tmpfile();
-			get_data_func(handle, fd, lilv_node_as_string(p->plugin_uri));
+			get_data_func(p->dynmanifest->handle, fd,
+			              lilv_node_as_string(p->plugin_uri));
 			rewind(fd);
-			serd_reader_read_file_handle(reader, fd,
-			                             (const uint8_t*)"(dyn-manifest)");
+			serd_reader_add_blank_prefix(
+				reader, lilv_world_blank_node_prefix(p->world));
+			serd_reader_read_file_handle(
+				reader, fd, (const uint8_t*)"(dyn-manifest)");
 			fclose(fd);
 		}
-
-		typedef int (*CloseFunc)(LV2_Dyn_Manifest_Handle);
-		CloseFunc close_func = (CloseFunc)lilv_dlfunc(lib, "lv2_dyn_manifest_close");
-		if (close_func)
-			close_func(handle);
 	}
 #endif
 	serd_reader_free(reader);
@@ -181,6 +181,21 @@ lilv_plugin_load(LilvPlugin* p)
 	p->loaded = true;
 }
 
+static bool
+is_symbol(const char* str)
+{
+	const size_t len = strlen(str);
+	for (size_t i = 0; i < len; ++i) {
+		if (!((str[i] >= 'a' && str[i] <= 'z')
+		      || (str[i] >= 'A' && str[i] <= 'Z')
+		      || (i > 0 && str[i] >= '0' && str[i] <= '9')
+		      || str[i] == '_')) {
+			return false;
+		}
+	}
+	return true;
+}
+
 static void
 lilv_plugin_load_ports_if_necessary(const LilvPlugin* const_p)
 {
@@ -189,35 +204,36 @@ lilv_plugin_load_ports_if_necessary(const LilvPlugin* const_p)
 		lilv_plugin_load(p);
 
 	if (!p->ports) {
-		p->ports = malloc(sizeof(LilvPort*));
+		p->ports = (LilvPort**)malloc(sizeof(LilvPort*));
 		p->ports[0] = NULL;
 
 		SordIter* ports = lilv_world_query_internal(
 			p->world,
 			p->plugin_uri->val.uri_val,
-			p->world->lv2_port_node,
+			p->world->uris.lv2_port,
 			NULL);
 
 		FOREACH_MATCH(ports) {
-			LilvNode*       index  = NULL;
-			const SordNode* port   = lilv_match_object(ports);
-			LilvNode*       symbol = lilv_plugin_get_unique(
-				p, port, p->world->lv2_symbol_node);
+			const SordNode* port = sord_iter_get_node(ports, SORD_OBJECT);
+			LilvNode* index = lilv_plugin_get_unique(
+				p, port, p->world->uris.lv2_index);
+			LilvNode* symbol = lilv_plugin_get_unique(
+				p, port, p->world->uris.lv2_symbol);
 
-			if (!lilv_node_is_string(symbol)) {
-				LILV_ERRORF("Plugin <%s> port symbol is not a string\n",
-				            lilv_node_as_uri(p->plugin_uri));
-				p->num_ports = 0;
-				goto error;
+			bool error = false;
+			if (!lilv_node_is_string(symbol) || !is_symbol(symbol->str_val)) {
+				LILV_ERRORF("Plugin <%s> port symbol `%s' is invalid\n",
+				            lilv_node_as_uri(p->plugin_uri),
+				            lilv_node_as_string(symbol));
+				error = true;
+				goto done;
 			}
 
-			index = lilv_plugin_get_unique(p, port, p->world->lv2_index_node);
-
 			if (!lilv_node_is_int(index)) {
 				LILV_ERRORF("Plugin <%s> port index is not an integer\n",
 				            lilv_node_as_uri(p->plugin_uri));
-				p->num_ports = 0;
-				goto error;
+				error = true;
+				goto done;
 			}
 
 			uint32_t  this_index = lilv_node_as_int(index);
@@ -225,7 +241,8 @@ lilv_plugin_load_ports_if_necessary(const LilvPlugin* const_p)
 			if (p->num_ports > this_index) {
 				this_port = p->ports[this_index];
 			} else {
-				p->ports = realloc(p->ports, (this_index + 1) * sizeof(LilvPort*));
+				p->ports = (LilvPort**)realloc(
+					p->ports, (this_index + 1) * sizeof(LilvPort*));
 				memset(p->ports + p->num_ports, '\0',
 				       (this_index - p->num_ports) * sizeof(LilvPort*));
 				p->num_ports = this_index + 1;
@@ -241,35 +258,39 @@ lilv_plugin_load_ports_if_necessary(const LilvPlugin* const_p)
 			}
 
 			SordIter* types = lilv_world_query_internal(
-				p->world, port, p->world->rdf_a_node, NULL);
+				p->world, port, p->world->uris.rdf_a, NULL);
 			FOREACH_MATCH(types) {
-				const SordNode* type = lilv_match_object(types);
+				const SordNode* type = sord_iter_get_node(types, SORD_OBJECT);
 				if (sord_node_get_type(type) == SORD_URI) {
 					zix_tree_insert(
-						this_port->classes,
+						(ZixTree*)this_port->classes,
 						lilv_node_new_from_node(p->world, type), NULL);
 				} else {
 					LILV_WARNF("Plugin <%s> port type is not a URI\n",
 					           lilv_node_as_uri(p->plugin_uri));
 				}
 			}
-			lilv_match_end(types);
+			sord_iter_free(types);
 
-		error:
+		done:
 			lilv_node_free(symbol);
 			lilv_node_free(index);
-			if (p->num_ports == 0) {
-				if (p->ports) {
-					for (uint32_t i = 0; i < p->num_ports; ++i) {
-						lilv_port_free(p, p->ports[i]);
-					}
-					free(p->ports);
-					p->ports = NULL;
-				}
-				break;  // Invalid plugin
+			if (error) {  // Invalid plugin
+				lilv_plugin_free_ports(p);
+				break;
+			}
+		}
+		sord_iter_free(ports);
+
+		// Check sanity
+		for (uint32_t i = 0; i < p->num_ports; ++i) {
+			if (!p->ports[i]) {
+				LILV_ERRORF("Plugin <%s> is missing port %d/%d\n",
+				            lilv_node_as_uri(p->plugin_uri), i, p->num_ports);
+				lilv_plugin_free_ports(p);
+				break;
 			}
 		}
-		lilv_match_end(ports);
 	}
 }
 
@@ -309,16 +330,16 @@ lilv_plugin_get_library_uri(const LilvPlugin* const_p)
 		SordIter* results = lilv_world_query_internal(
 			p->world,
 			p->plugin_uri->val.uri_val,
-			p->world->lv2_binary_node,
+			p->world->uris.lv2_binary,
 			NULL);
 		FOREACH_MATCH(results) {
-			const SordNode* binary_node = lilv_match_object(results);
+			const SordNode* binary_node = sord_iter_get_node(results, SORD_OBJECT);
 			if (sord_node_get_type(binary_node) == SORD_URI) {
 				p->binary_uri = lilv_node_new_from_node(p->world, binary_node);
 				break;
 			}
 		}
-		lilv_match_end(results);
+		sord_iter_free(results);
 	}
 	if (!p->binary_uri) {
 		LILV_WARNF("Plugin <%s> has no lv2:binary\n",
@@ -345,29 +366,29 @@ lilv_plugin_get_class(const LilvPlugin* const_p)
 		SordIter* results = lilv_world_query_internal(
 			p->world,
 			p->plugin_uri->val.uri_val,
-			p->world->rdf_a_node,
+			p->world->uris.rdf_a,
 			NULL);
 		FOREACH_MATCH(results) {
-			const SordNode* class_node = lilv_match_object(results);
+			const SordNode* class_node = sord_iter_get_node(results, SORD_OBJECT);
 			if (sord_node_get_type(class_node) != SORD_URI) {
 				continue;
 			}
 
-			LilvNode* class = lilv_node_new_from_node(p->world, class_node);
-			if (!lilv_node_equals(class, p->world->lv2_plugin_class->uri)) {
-				const LilvPluginClass* plugin_class = lilv_plugin_classes_get_by_uri(
-					p->world->plugin_classes, class);
+			LilvNode* klass = lilv_node_new_from_node(p->world, class_node);
+			if (!lilv_node_equals(klass, p->world->lv2_plugin_class->uri)) {
+				const LilvPluginClass* pclass = lilv_plugin_classes_get_by_uri(
+					p->world->plugin_classes, klass);
 
-				if (plugin_class) {
-					((LilvPlugin*)p)->plugin_class = plugin_class;
-					lilv_node_free(class);
+				if (pclass) {
+					((LilvPlugin*)p)->plugin_class = pclass;
+					lilv_node_free(klass);
 					break;
 				}
 			}
 
-			lilv_node_free(class);
+			lilv_node_free(klass);
 		}
-		lilv_match_end(results);
+		sord_iter_free(results);
 
 		if (p->plugin_class == NULL)
 			p->plugin_class = p->world->lv2_plugin_class;
@@ -375,6 +396,15 @@ lilv_plugin_get_class(const LilvPlugin* const_p)
 	return p->plugin_class;
 }
 
+static LilvNodes*
+lilv_plugin_get_value_internal(const LilvPlugin* p,
+                               const SordNode*   predicate)
+{
+	lilv_plugin_load_if_necessary(p);
+	return lilv_world_query_values_internal(
+		p->world, p->plugin_uri->val.uri_val, predicate, NULL);
+}
+
 LILV_API
 bool
 lilv_plugin_verify(const LilvPlugin* plugin)
@@ -387,13 +417,14 @@ lilv_plugin_verify(const LilvPlugin* plugin)
 	}
 
 	lilv_nodes_free(results);
-	results = lilv_plugin_get_value(plugin, plugin->world->doap_name_val);
+	results = lilv_plugin_get_value_internal(plugin,
+	                                         plugin->world->uris.doap_name);
 	if (!results) {
 		return false;
 	}
 
 	lilv_nodes_free(results);
-	LilvNode*  lv2_port = lilv_new_uri(plugin->world, LILV_NS_LV2 "port");
+	LilvNode* lv2_port = lilv_new_uri(plugin->world, LV2_CORE__port);
 	results = lilv_plugin_get_value(plugin, lv2_port);
 	lilv_node_free(lv2_port);
 	if (!results) {
@@ -408,8 +439,8 @@ LILV_API
 LilvNode*
 lilv_plugin_get_name(const LilvPlugin* plugin)
 {
-	LilvNodes* results = lilv_plugin_get_value(plugin,
-	                                           plugin->world->doap_name_val);
+	LilvNodes* results = lilv_plugin_get_value_internal(
+		plugin, plugin->world->uris.doap_name);
 
 	LilvNode* ret = NULL;
 	if (results) {
@@ -462,30 +493,24 @@ lilv_plugin_get_port_ranges_float(const LilvPlugin* p,
 		lilv_port_get_range(p, p->ports[i], defptr, minptr, maxptr);
 
 		if (min_values) {
-			if (lilv_node_is_float(min)) {
+			if (lilv_node_is_float(min) || lilv_node_is_int(min)) {
 				min_values[i] = lilv_node_as_float(min);
-			} else if (lilv_node_is_int(min)) {
-				min_values[i] = lilv_node_as_int(min);
 			} else {
 				min_values[i] = NAN;
 			}
 		}
 
 		if (max_values) {
-			if (lilv_node_is_float(max)) {
+			if (lilv_node_is_float(max) || lilv_node_is_int(max)) {
 				max_values[i] = lilv_node_as_float(max);
-			} else if (lilv_node_is_int(max)) {
-				max_values[i] = lilv_node_as_int(max);
 			} else {
 				max_values[i] = NAN;
 			}
 		}
 
 		if (def_values) {
-			if (lilv_node_is_float(def)) {
+			if (lilv_node_is_float(def) || lilv_node_is_int(def)) {
 				def_values[i] = lilv_node_as_float(def);
-			} else if (lilv_node_is_int(def)) {
-				def_values[i] = lilv_node_as_int(def);
 			} else {
 				def_values[i] = NAN;
 			}
@@ -540,62 +565,90 @@ lilv_plugin_has_latency(const LilvPlugin* p)
 	SordIter* ports = lilv_world_query_internal(
 		p->world,
 		p->plugin_uri->val.uri_val,
-		p->world->lv2_port_node,
+		p->world->uris.lv2_port,
 		NULL);
 
 	bool ret = false;
 	FOREACH_MATCH(ports) {
-		const SordNode* port            = lilv_match_object(ports);
+		const SordNode* port            = sord_iter_get_node(ports, SORD_OBJECT);
 		SordIter*       reports_latency = lilv_world_query_internal(
 			p->world,
 			port,
-			p->world->lv2_portproperty_node,
-			p->world->lv2_reportslatency_node);
-		const bool end = lilv_matches_end(reports_latency);
-		lilv_match_end(reports_latency);
+			p->world->uris.lv2_portProperty,
+			p->world->uris.lv2_reportsLatency);
+		const bool end = sord_iter_end(reports_latency);
+		sord_iter_free(reports_latency);
 		if (!end) {
 			ret = true;
 			break;
 		}
 	}
-	lilv_match_end(ports);
+	sord_iter_free(ports);
 
 	return ret;
 }
 
+static LilvPort*
+lilv_plugin_get_port_by_property(const LilvPlugin* plugin,
+                                 const LilvNode*   port_property)
+{
+	lilv_plugin_load_ports_if_necessary(plugin);
+	for (uint32_t i = 0; i < plugin->num_ports; ++i) {
+		LilvPort* port = plugin->ports[i];
+		SordIter* iter = lilv_world_query_internal(
+			plugin->world,
+			port->node,
+			plugin->world->uris.lv2_portProperty,
+			port_property->val.uri_val);
+
+		const bool found = !sord_iter_end(iter);
+		sord_iter_free(iter);
+
+		if (found) {
+			return port;
+		}
+	}
+
+	return NULL;
+}
+
+LILV_API
+LilvPort*
+lilv_plugin_get_port_by_designation(const LilvPlugin* plugin,
+                                    const LilvNode*   port_class,
+                                    const LilvNode*   designation)
+{
+	LilvWorld* world = plugin->world;
+	lilv_plugin_load_ports_if_necessary(plugin);
+	for (uint32_t i = 0; i < plugin->num_ports; ++i) {
+		LilvPort* port = plugin->ports[i];
+		SordIter* iter = lilv_world_query_internal(
+			world,
+			port->node,
+			world->uris.lv2_designation,
+			designation->val.uri_val);
+
+		const bool found = !sord_iter_end(iter) &&
+			lilv_port_is_a(plugin, port, port_class);
+		sord_iter_free(iter);
+
+		if (found) {
+			return port;
+		}
+	}
+
+	return NULL;
+}
+
 LILV_API
 uint32_t
 lilv_plugin_get_latency_port_index(const LilvPlugin* p)
 {
-	lilv_plugin_load_if_necessary(p);
-	SordIter* ports = lilv_world_query_internal(
-		p->world,
-		p->plugin_uri->val.uri_val,
-		p->world->lv2_port_node,
-		NULL);
-
-	uint32_t ret = 0;
-	FOREACH_MATCH(ports) {
-		const SordNode* port            = lilv_match_object(ports);
-		SordIter*       reports_latency = lilv_world_query_internal(
-			p->world,
-			port,
-			p->world->lv2_portproperty_node,
-			p->world->lv2_reportslatency_node);
-		if (!lilv_matches_end(reports_latency)) {
-			LilvNode* index = lilv_plugin_get_unique(
-				p, port, p->world->lv2_index_node);
-
-			ret = lilv_node_as_int(index);
-			lilv_node_free(index);
-			lilv_match_end(reports_latency);
-			break;
-		}
-		lilv_match_end(reports_latency);
-	}
-	lilv_match_end(ports);
-
-	return ret;  // FIXME: error handling
+	LilvNode* property = lilv_node_new_from_node(
+		p->world, p->world->uris.lv2_reportsLatency);
+	LilvPort* latency_port = lilv_plugin_get_port_by_property(p, property);
+	lilv_node_free(property);
+	return latency_port ? latency_port->index : UINT32_MAX;
 }
 
 LILV_API
@@ -605,7 +658,8 @@ lilv_plugin_has_feature(const LilvPlugin* p,
 {
 	LilvNodes* features = lilv_plugin_get_supported_features(p);
 
-	const bool ret = features && feature && lilv_nodes_contains(features, feature);
+	const bool ret = (features && feature
+	                  && lilv_nodes_contains(features, feature));
 
 	lilv_nodes_free(features);
 	return ret;
@@ -620,11 +674,13 @@ lilv_plugin_get_supported_features(const LilvPlugin* p)
 	LilvNodes* result   = lilv_nodes_new();
 
 	LILV_FOREACH(nodes, i, optional)
-		zix_tree_insert(
-			result, lilv_node_duplicate(lilv_nodes_get(optional, i)), NULL);
+		zix_tree_insert((ZixTree*)result,
+		                lilv_node_duplicate(lilv_nodes_get(optional, i)),
+		                NULL);
 	LILV_FOREACH(nodes, i, required)
-		zix_tree_insert(
-			result, lilv_node_duplicate(lilv_nodes_get(required, i)), NULL);
+		zix_tree_insert((ZixTree*)result,
+		                lilv_node_duplicate(lilv_nodes_get(required, i)),
+		                NULL);
 
 	lilv_nodes_free(optional);
 	lilv_nodes_free(required);
@@ -636,14 +692,47 @@ LILV_API
 LilvNodes*
 lilv_plugin_get_optional_features(const LilvPlugin* p)
 {
-	return lilv_plugin_get_value(p, p->world->lv2_optionalFeature_val);
+	return lilv_plugin_get_value_internal(
+		p, p->world->uris.lv2_optionalFeature);
 }
 
 LILV_API
 LilvNodes*
 lilv_plugin_get_required_features(const LilvPlugin* p)
 {
-	return lilv_plugin_get_value(p, p->world->lv2_requiredFeature_val);
+	return lilv_plugin_get_value_internal(
+		p, p->world->uris.lv2_requiredFeature);
+}
+
+LILV_API
+bool
+lilv_plugin_has_extension_data(const LilvPlugin* p,
+                               const LilvNode*   uri)
+{
+	if (!lilv_node_is_uri(uri)) {
+		LILV_ERRORF("Extension data `%s' is not a URI\n", uri->str_val);
+		return false;
+	}
+
+	SordIter* iter = lilv_world_query_internal(
+		p->world,
+		p->plugin_uri->val.uri_val,
+		p->world->uris.lv2_extensionData,
+		uri->val.uri_val);
+
+	if (iter) {
+		sord_iter_free(iter);
+		return true;
+	} else {
+		return false;
+	}
+}
+
+LILV_API
+LilvNodes*
+lilv_plugin_get_extension_data(const LilvPlugin* p)
+{
+	return lilv_plugin_get_value_internal(p, p->world->uris.lv2_extensionData);
 }
 
 LILV_API
@@ -673,6 +762,33 @@ lilv_plugin_get_port_by_symbol(const LilvPlugin* p,
 	return NULL;
 }
 
+LILV_API
+LilvNode*
+lilv_plugin_get_project(const LilvPlugin* p)
+{
+	lilv_plugin_load_if_necessary(p);
+
+	SordNode* lv2_project = sord_new_uri(p->world->world,
+	                                     (const uint8_t*)LV2_CORE__project);
+
+	SordIter* projects = lilv_world_query_internal(
+		p->world,
+		p->plugin_uri->val.uri_val,
+		lv2_project,
+		NULL);
+
+	sord_node_free(p->world->world, lv2_project);
+
+	if (sord_iter_end(projects)) {
+		return NULL;
+	}
+
+	const SordNode* project = sord_iter_get_node(projects, SORD_OBJECT);
+
+	sord_iter_free(projects);
+	return lilv_node_new_from_node(p->world, project);
+}
+
 static const SordNode*
 lilv_plugin_get_author(const LilvPlugin* p)
 {
@@ -689,13 +805,26 @@ lilv_plugin_get_author(const LilvPlugin* p)
 
 	sord_node_free(p->world->world, doap_maintainer);
 
-	if (lilv_matches_end(maintainers)) {
+	if (sord_iter_end(maintainers)) {
+		LilvNode* project = lilv_plugin_get_project(p);
+		if (!project) {
+			return NULL;
+		}
+
+		maintainers = lilv_world_query_internal(
+			p->world,
+			project->val.uri_val,
+			doap_maintainer,
+			NULL);
+	}
+
+	if (sord_iter_end(maintainers)) {
 		return NULL;
 	}
 
-	const SordNode* author = lilv_match_object(maintainers);
+	const SordNode* author = sord_iter_get_node(maintainers, SORD_OBJECT);
 
-	lilv_match_end(maintainers);
+	sord_iter_free(maintainers);
 	return author;
 }
 
@@ -751,8 +880,10 @@ lilv_plugin_get_uis(const LilvPlugin* p)
 {
 	lilv_plugin_load_if_necessary(p);
 
-	SordNode* ui_ui_node     = sord_new_uri(p->world->world, NS_UI "ui");
-	SordNode* ui_binary_node = sord_new_uri(p->world->world, NS_UI "binary");
+	SordNode* ui_ui_node = sord_new_uri(p->world->world,
+	                                    (const uint8_t*)LV2_UI__ui);
+	SordNode* ui_binary_node = sord_new_uri(p->world->world,
+	                                        (const uint8_t*)LV2_UI__binary);
 
 	LilvUIs*  result = lilv_uis_new();
 	SordIter* uis    = lilv_world_query_internal(
@@ -762,9 +893,10 @@ lilv_plugin_get_uis(const LilvPlugin* p)
 		NULL);
 
 	FOREACH_MATCH(uis) {
-		const SordNode* ui     = lilv_match_object(uis);
-		LilvNode*       type   = lilv_plugin_get_unique(p, ui, p->world->rdf_a_node);
-		LilvNode*       binary = lilv_plugin_get_unique(p, ui, ui_binary_node);
+		const SordNode* ui = sord_iter_get_node(uis, SORD_OBJECT);
+
+		LilvNode* type   = lilv_plugin_get_unique(p, ui, p->world->uris.rdf_a);
+		LilvNode* binary = lilv_plugin_get_unique(p, ui, ui_binary_node);
 
 		if (sord_node_get_type(ui) != SORD_URI
 		    || !lilv_node_is_uri(type)
@@ -781,9 +913,9 @@ lilv_plugin_get_uis(const LilvPlugin* p)
 			type,
 			binary);
 
-		zix_tree_insert(result, lilv_ui, NULL);
+		zix_tree_insert((ZixTree*)result, lilv_ui, NULL);
 	}
-	lilv_match_end(uis);
+	sord_iter_free(uis);
 
 	sord_node_free(p->world->world, ui_binary_node);
 	sord_node_free(p->world->world, ui_ui_node);
@@ -796,11 +928,36 @@ lilv_plugin_get_uis(const LilvPlugin* p)
 	}
 }
 
-static size_t
-file_sink(const void* buf, size_t len, void* stream)
+LILV_API
+LilvNodes*
+lilv_plugin_get_related(const LilvPlugin* plugin, const LilvNode* type)
 {
-	FILE* file = (FILE*)stream;
-	return fwrite(buf, 1, len, file);
+	LilvWorld* const world   = plugin->world;
+	LilvNodes* const related = lilv_world_query_values_internal(
+		world,
+		NULL,
+		world->uris.lv2_appliesTo,
+		lilv_plugin_get_uri(plugin)->val.uri_val);
+
+	if (!type) {
+		return related;
+	}
+
+	LilvNodes* matches = lilv_nodes_new();
+	LILV_FOREACH(nodes, i, related) {
+		LilvNode* node  = (LilvNode*)lilv_collection_get((ZixTree*)related, i);
+		SordIter* titer = lilv_world_query_internal(
+			world, node->val.uri_val, world->uris.rdf_a, type->val.uri_val);
+		if (!sord_iter_end(titer)) {
+			zix_tree_insert((ZixTree*)matches,
+			                lilv_node_new_from_node(world, node->val.uri_val),
+			                NULL);
+		}
+		sord_iter_free(titer);
+	}
+
+	lilv_nodes_free(related);
+	return matches;
 }
 
 static SerdEnv*
@@ -846,10 +1003,10 @@ lilv_plugin_write_description(LilvWorld*        world,
 
 	SerdWriter* writer = serd_writer_new(
 		SERD_TURTLE,
-		SERD_STYLE_ABBREVIATED|SERD_STYLE_CURIED,
+		(SerdStyle)(SERD_STYLE_ABBREVIATED|SERD_STYLE_CURIED),
 		env,
 		NULL,
-		file_sink,
+		serd_file_sink,
 		plugin_file);
 
 	// Write prefixes if this is a new file
@@ -886,10 +1043,10 @@ lilv_plugin_write_manifest_entry(LilvWorld*        world,
 
 	SerdWriter* writer = serd_writer_new(
 		SERD_TURTLE,
-		SERD_STYLE_ABBREVIATED|SERD_STYLE_CURIED,
+		(SerdStyle)(SERD_STYLE_ABBREVIATED|SERD_STYLE_CURIED),
 		env,
 		NULL,
-		file_sink,
+		serd_file_sink,
 		manifest_file);
 
 	// Write prefixes if this is a new file
@@ -899,15 +1056,15 @@ lilv_plugin_write_manifest_entry(LilvWorld*        world,
 	serd_writer_write_statement(
 		writer, 0, NULL,
 		sord_node_to_serd_node(subject->val.uri_val),
-		sord_node_to_serd_node(plugin->world->rdf_a_node),
-		sord_node_to_serd_node(plugin->world->lv2_plugin_node), 0, 0);
+		sord_node_to_serd_node(plugin->world->uris.rdf_a),
+		sord_node_to_serd_node(plugin->world->uris.lv2_Plugin), 0, 0);
 
 	const SerdNode file_node = serd_node_from_string(
 		SERD_URI, (const uint8_t*)plugin_file_path);
 	serd_writer_write_statement(
 		writer, 0, NULL,
 		sord_node_to_serd_node(subject->val.uri_val),
-		sord_node_to_serd_node(plugin->world->rdfs_seealso_node),
+		sord_node_to_serd_node(plugin->world->uris.rdfs_seeAlso),
 		&file_node, 0, 0);
 
 	serd_writer_free(writer);
