@@ -14,14 +14,14 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#define _XOPEN_SOURCE 500
-
 #include <assert.h>
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
+#include "lv2/lv2plug.in/ns/ext/event/event.h"
+
 #include "lilv_internal.h"
 
 LilvPort*
@@ -30,7 +30,7 @@ lilv_port_new(LilvWorld*      world,
               uint32_t        index,
               const char*     symbol)
 {
-	LilvPort* port = malloc(sizeof(struct LilvPortImpl));
+	LilvPort* port = (LilvPort*)malloc(sizeof(LilvPort));
 	port->node    = sord_node_copy(node);
 	port->index   = index;
 	port->symbol  = lilv_node_new(world, LILV_VALUE_STRING, symbol);
@@ -41,10 +41,12 @@ lilv_port_new(LilvWorld*      world,
 void
 lilv_port_free(const LilvPlugin* plugin, LilvPort* port)
 {
-	sord_node_free(plugin->world->world, port->node);
-	lilv_nodes_free(port->classes);
-	lilv_node_free(port->symbol);
-	free(port);
+	if (port) {
+		sord_node_free(plugin->world->world, port->node);
+		lilv_nodes_free(port->classes);
+		lilv_node_free(port->symbol);
+		free(port);
+	}
 }
 
 LILV_API
@@ -70,11 +72,11 @@ lilv_port_has_property(const LilvPlugin* p,
 	SordIter* results = lilv_world_query_internal(
 		p->world,
 		port->node,
-		p->world->lv2_portproperty_node,
+		p->world->uris.lv2_portProperty,
 		lilv_node_as_node(property));
 
-	const bool ret = !lilv_matches_end(results);
-	lilv_match_end(results);
+	const bool ret = !sord_iter_end(results);
+	sord_iter_free(results);
 	return ret;
 }
 
@@ -84,17 +86,15 @@ lilv_port_supports_event(const LilvPlugin* p,
                          const LilvPort*   port,
                          const LilvNode*   event)
 {
-#define NS_EV (const uint8_t*)"http://lv2plug.in/ns/ext/event#"
-
 	assert(event);
 	SordIter* results = lilv_world_query_internal(
 		p->world,
 		port->node,
-		sord_new_uri(p->world->world, NS_EV "supportsEvent"),
+		sord_new_uri(p->world->world, (const uint8_t*)LV2_EVENT__supportsEvent),
 		lilv_node_as_node(event));
 
-	const bool ret = !lilv_matches_end(results);
-	lilv_match_end(results);
+	const bool ret = !sord_iter_end(results);
+	sord_iter_free(results);
 	return ret;
 }
 
@@ -111,7 +111,7 @@ lilv_port_get_value_by_node(const LilvPlugin* p,
 		predicate,
 		NULL);
 
-	return lilv_nodes_from_stream_objects(p->world, results);
+	return lilv_nodes_from_stream_objects(p->world, results, SORD_OBJECT);
 }
 
 LILV_API
@@ -130,6 +130,14 @@ lilv_port_get_value(const LilvPlugin* p,
 		lilv_node_as_node(predicate));
 }
 
+LILV_API
+uint32_t
+lilv_port_get_index(const LilvPlugin* p,
+                    const LilvPort*   port)
+{
+	return port->index;
+}
+
 LILV_API
 const LilvNode*
 lilv_port_get_symbol(const LilvPlugin* p,
@@ -143,8 +151,8 @@ LilvNode*
 lilv_port_get_name(const LilvPlugin* p,
                    const LilvPort*   port)
 {
-	LilvNodes* results = lilv_port_get_value(p, port,
-	                                         p->world->lv2_name_val);
+	LilvNodes* results = lilv_port_get_value_by_node(
+		p, port, p->world->uris.lv2_name);
 
 	LilvNode* ret = NULL;
 	if (results) {
@@ -179,7 +187,7 @@ lilv_port_get_range(const LilvPlugin* p,
 {
 	if (def) {
 		LilvNodes* defaults = lilv_port_get_value_by_node(
-			p, port, p->world->lv2_default_node);
+			p, port, p->world->uris.lv2_default);
 		*def = defaults
 			? lilv_node_duplicate(lilv_nodes_get_first(defaults))
 			: NULL;
@@ -187,7 +195,7 @@ lilv_port_get_range(const LilvPlugin* p,
 	}
 	if (min) {
 		LilvNodes* minimums = lilv_port_get_value_by_node(
-			p, port, p->world->lv2_minimum_node);
+			p, port, p->world->uris.lv2_minimum);
 		*min = minimums
 			? lilv_node_duplicate(lilv_nodes_get_first(minimums))
 			: NULL;
@@ -195,7 +203,7 @@ lilv_port_get_range(const LilvPlugin* p,
 	}
 	if (max) {
 		LilvNodes* maximums = lilv_port_get_value_by_node(
-			p, port, p->world->lv2_maximum_node);
+			p, port, p->world->uris.lv2_maximum);
 		*max = maximums
 			? lilv_node_duplicate(lilv_nodes_get_first(maximums))
 			: NULL;
@@ -211,31 +219,32 @@ lilv_port_get_scale_points(const LilvPlugin* p,
 	SordIter* points = lilv_world_query_internal(
 		p->world,
 		port->node,
-		sord_new_uri(p->world->world, (const uint8_t*)LILV_NS_LV2 "scalePoint"),
+		sord_new_uri(p->world->world, (const uint8_t*)LV2_CORE__scalePoint),
 		NULL);
 
 	LilvScalePoints* ret = NULL;
-	if (!lilv_matches_end(points))
+	if (!sord_iter_end(points))
 		ret = lilv_scale_points_new();
 
 	FOREACH_MATCH(points) {
-		const SordNode* point = lilv_match_object(points);
+		const SordNode* point = sord_iter_get_node(points, SORD_OBJECT);
 
 		LilvNode* value = lilv_plugin_get_unique(
 			p,
 			point,
-			p->world->rdf_value_node);
+			p->world->uris.rdf_value);
 
 		LilvNode* label = lilv_plugin_get_unique(
 			p,
 			point,
-			p->world->rdfs_label_node);
+			p->world->uris.rdfs_label);
 
 		if (value && label) {
-			zix_tree_insert(ret, lilv_scale_point_new(value, label), NULL);
+			zix_tree_insert(
+				(ZixTree*)ret, lilv_scale_point_new(value, label), NULL);
 		}
 	}
-	lilv_match_end(points);
+	sord_iter_free(points);
 
 	assert(!ret || lilv_nodes_size(ret) > 0);
 	return ret;
@@ -247,7 +256,7 @@ lilv_port_get_properties(const LilvPlugin* p,
                          const LilvPort*   port)
 {
 	LilvNode* pred = lilv_node_new_from_node(
-		p->world, p->world->lv2_portproperty_node);
+		p->world, p->world->uris.lv2_portProperty);
 	LilvNodes* ret = lilv_port_get_value(p, port, pred);
 	lilv_node_free(pred);
 	return ret;
