@@ -14,8 +14,6 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#define _XOPEN_SOURCE 500
-
 #include <assert.h>
 #include <limits.h>
 #include <stdlib.h>
@@ -37,32 +35,28 @@ lilv_lang_matches(const char* a, const char* b)
 	}
 
 	const char*  a_dash     = strchr(a, '-');
-	const size_t a_lang_len = a_dash ? (a_dash - a) : 0;
+	const size_t a_lang_len = a_dash ? (size_t)(a_dash - a) : strlen(a);
 	const char*  b_dash     = strchr(b, '-');
-	const size_t b_lang_len = b_dash ? (b_dash - b) : 0;
+	const size_t b_lang_len = b_dash ? (size_t)(b_dash - b) : strlen(b);
 
-	if (a_lang_len && b_lang_len) {
-		if (a_lang_len == b_lang_len && !strncmp(a, b, a_lang_len)) {
-			return LILV_LANG_MATCH_PARTIAL;  // e.g. a="en-gb", b="en-ca"
-		}
-	} else if (a_lang_len && !strncmp(a, b, a_lang_len)) {
-		return LILV_LANG_MATCH_PARTIAL;  // e.g. a="en", b="en-ca"
-	} else if (b_lang_len && !strncmp(a, b, b_lang_len)) {
-		return LILV_LANG_MATCH_PARTIAL;  // e.g. a="en-ca", b="en"
+	if (a_lang_len == b_lang_len && !strncmp(a, b, a_lang_len)) {
+		return LILV_LANG_MATCH_PARTIAL;
 	}
+
 	return LILV_LANG_MATCH_NONE;
 }
 
 LilvNodes*
-lilv_nodes_from_stream_objects_i18n(LilvWorld* world,
-                                    SordIter*  stream)
+lilv_nodes_from_stream_objects_i18n(LilvWorld*    world,
+                                    SordIter*     stream,
+                                    SordQuadIndex field)
 {
 	LilvNodes*      values  = lilv_nodes_new();
 	const SordNode* nolang  = NULL;  // Untranslated value
 	const SordNode* partial = NULL;  // Partial language match
 	char*           syslang = lilv_get_lang();
 	FOREACH_MATCH(stream) {
-		const SordNode* value = lilv_match_object(stream);
+		const SordNode* value = sord_iter_get_node(stream, field);
 		if (sord_node_get_type(value) == SORD_LITERAL) {
 			const char*   lang = sord_node_get_language(value);
 			LilvLangMatch lm   = LILV_LANG_MATCH_NONE;
@@ -79,17 +73,20 @@ lilv_nodes_from_stream_objects_i18n(LilvWorld* world,
 
 			if (lm == LILV_LANG_MATCH_EXACT) {
 				// Exact language match, add to results
-				zix_tree_insert(
-					values, lilv_node_new_from_node(world, value), NULL);
+				zix_tree_insert((ZixTree*)values,
+				                lilv_node_new_from_node(world, value),
+				                NULL);
 			} else if (lm == LILV_LANG_MATCH_PARTIAL) {
 				// Partial language match, save in case we find no exact
 				partial = value;
 			}
 		} else {
-			zix_tree_insert(values, lilv_node_new_from_node(world, value), NULL);
+			zix_tree_insert((ZixTree*)values,
+			                lilv_node_new_from_node(world, value),
+			                NULL);
 		}
 	}
-	lilv_match_end(stream);
+	sord_iter_free(stream);
 	free(syslang);
 
 	if (lilv_nodes_size(values) > 0) {
@@ -107,7 +104,8 @@ lilv_nodes_from_stream_objects_i18n(LilvWorld* world,
 	}
 
 	if (best) {
-		zix_tree_insert(values, lilv_node_new_from_node(world, best), NULL);
+		zix_tree_insert(
+			(ZixTree*)values, lilv_node_new_from_node(world, best), NULL);
 	} else {
 		// No matches whatsoever
 		lilv_nodes_free(values);
@@ -118,24 +116,25 @@ lilv_nodes_from_stream_objects_i18n(LilvWorld* world,
 }
 
 LilvNodes*
-lilv_nodes_from_stream_objects(LilvWorld* world,
-                               SordIter*  stream)
+lilv_nodes_from_stream_objects(LilvWorld*    world,
+                               SordIter*     stream,
+                               SordQuadIndex field)
 {
-	if (lilv_matches_end(stream)) {
-		lilv_match_end(stream);
+	if (sord_iter_end(stream)) {
+		sord_iter_free(stream);
 		return NULL;
 	} else if (world->opt.filter_language) {
-		return lilv_nodes_from_stream_objects_i18n(world, stream);
+		return lilv_nodes_from_stream_objects_i18n(world, stream, field);
 	} else {
 		LilvNodes* values = lilv_nodes_new();
 		FOREACH_MATCH(stream) {
-			LilvNode* value = lilv_node_new_from_node(
-				world, lilv_match_object(stream));
-			if (value) {
-				zix_tree_insert(values, value, NULL);
+			const SordNode* value = sord_iter_get_node(stream, field);
+			LilvNode* node = lilv_node_new_from_node(world, value);
+			if (node) {
+				zix_tree_insert((ZixTree*)values, node, NULL);
 			}
 		}
-		lilv_match_end(stream);
+		sord_iter_free(stream);
 		return values;
 	}
 }
