@@ -14,17 +14,12 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#define _XOPEN_SOURCE 500
-
 #include <assert.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include <dirent.h>
-#ifdef HAVE_WORDEXP
-#include <wordexp.h>
-#endif
+#include "lv2/lv2plug.in/ns/ext/presets/presets.h"
 
 #include "lilv_internal.h"
 
@@ -32,7 +27,7 @@ LILV_API
 LilvWorld*
 lilv_world_new(void)
 {
-	LilvWorld* world = malloc(sizeof(struct LilvWorldImpl));
+	LilvWorld* world = (LilvWorld*)malloc(sizeof(LilvWorld));
 
 	world->world = sord_world_new();
 	if (!world->world)
@@ -45,44 +40,55 @@ lilv_world_new(void)
 	world->specs          = NULL;
 	world->plugin_classes = lilv_plugin_classes_new();
 	world->plugins        = lilv_plugins_new();
+	world->loaded_files   = zix_tree_new(
+		false, lilv_resource_node_cmp, NULL, (ZixDestroyFunc)lilv_node_free);
 
-#define NS_DYNMAN "http://lv2plug.in/ns/ext/dynmanifest#"
-#define NS_DC     "http://dublincore.org/documents/dcmi-namespace/"
+#ifdef LILV_NEW_LV2
+	world->libs = zix_tree_new(
+		false, lilv_header_compare_by_uri, NULL, NULL);
+#endif
 
-#define NEW_URI(uri)     sord_new_uri(world->world, (const uint8_t*)uri)
-#define NEW_URI_VAL(uri) lilv_new_uri(world, (const char*)(uri));
+#define NS_DCTERMS "http://purl.org/dc/terms/"
+#define NS_DYNMAN  "http://lv2plug.in/ns/ext/dynmanifest#"
 
-	world->dc_replaces_node         = NEW_URI(NS_DC        "replaces");
-	world->dyn_manifest_node        = NEW_URI(NS_DYNMAN    "DynManifest");
-	world->lv2_binary_node          = NEW_URI(LILV_NS_LV2  "binary");
-	world->lv2_default_node         = NEW_URI(LILV_NS_LV2  "default");
-	world->lv2_index_node           = NEW_URI(LILV_NS_LV2  "index");
-	world->lv2_maximum_node         = NEW_URI(LILV_NS_LV2  "maximum");
-	world->lv2_minimum_node         = NEW_URI(LILV_NS_LV2  "minimum");
-	world->lv2_plugin_node          = NEW_URI(LILV_NS_LV2  "Plugin");
-	world->lv2_port_node            = NEW_URI(LILV_NS_LV2  "port");
-	world->lv2_portproperty_node    = NEW_URI(LILV_NS_LV2  "portProperty");
-	world->lv2_reportslatency_node  = NEW_URI(LILV_NS_LV2  "reportsLatency");
-	world->lv2_specification_node   = NEW_URI(LILV_NS_LV2  "Specification");
-	world->lv2_symbol_node          = NEW_URI(LILV_NS_LV2  "symbol");
-	world->rdf_a_node               = NEW_URI(LILV_NS_RDF  "type");
-	world->rdf_value_node           = NEW_URI(LILV_NS_RDF  "value");
-	world->rdfs_class_node          = NEW_URI(LILV_NS_RDFS "Class");
-	world->rdfs_label_node          = NEW_URI(LILV_NS_RDFS "label");
-	world->rdfs_seealso_node        = NEW_URI(LILV_NS_RDFS "seeAlso");
-	world->rdfs_subclassof_node     = NEW_URI(LILV_NS_RDFS "subClassOf");
-	world->xsd_boolean_node         = NEW_URI(LILV_NS_XSD  "boolean");
-	world->xsd_decimal_node         = NEW_URI(LILV_NS_XSD  "decimal");
-	world->xsd_double_node          = NEW_URI(LILV_NS_XSD  "double");
-	world->xsd_integer_node         = NEW_URI(LILV_NS_XSD  "integer");
+#define NEW_URI(uri) sord_new_uri(world->world, (const uint8_t*)uri)
 
-	world->doap_name_val           = NEW_URI_VAL(LILV_NS_DOAP "name");
-	world->lv2_name_val            = NEW_URI_VAL(LILV_NS_LV2  "name");
-	world->lv2_optionalFeature_val = NEW_URI_VAL(LILV_NS_LV2  "optionalFeature");
-	world->lv2_requiredFeature_val = NEW_URI_VAL(LILV_NS_LV2  "requiredFeature");
+	world->uris.dc_replaces         = NEW_URI(NS_DCTERMS   "replaces");
+	world->uris.dman_DynManifest    = NEW_URI(NS_DYNMAN    "DynManifest");
+	world->uris.doap_name           = NEW_URI(LILV_NS_DOAP "name");
+	world->uris.lv2_Plugin          = NEW_URI(LV2_CORE__Plugin);
+	world->uris.lv2_Specification   = NEW_URI(LV2_CORE__Specification);
+	world->uris.lv2_appliesTo       = NEW_URI(LV2_CORE__appliesTo);
+	world->uris.lv2_binary          = NEW_URI(LV2_CORE__binary);
+	world->uris.lv2_default         = NEW_URI(LV2_CORE__default);
+	world->uris.lv2_designation     = NEW_URI(LV2_CORE__designation);
+	world->uris.lv2_extensionData   = NEW_URI(LV2_CORE__extensionData);
+	world->uris.lv2_index           = NEW_URI(LV2_CORE__index);
+	world->uris.lv2_maximum         = NEW_URI(LV2_CORE__maximum);
+	world->uris.lv2_minimum         = NEW_URI(LV2_CORE__minimum);
+	world->uris.lv2_name            = NEW_URI(LV2_CORE__name);
+	world->uris.lv2_optionalFeature = NEW_URI(LV2_CORE__optionalFeature);
+	world->uris.lv2_port            = NEW_URI(LV2_CORE__port);
+	world->uris.lv2_portProperty    = NEW_URI(LV2_CORE__portProperty);
+	world->uris.lv2_reportsLatency  = NEW_URI(LV2_CORE__reportsLatency);
+	world->uris.lv2_requiredFeature = NEW_URI(LV2_CORE__requiredFeature);
+	world->uris.lv2_symbol          = NEW_URI(LV2_CORE__symbol);
+	world->uris.pset_value          = NEW_URI(LV2_PRESETS__value);
+	world->uris.rdf_a               = NEW_URI(LILV_NS_RDF  "type");
+	world->uris.rdf_value           = NEW_URI(LILV_NS_RDF  "value");
+	world->uris.rdfs_Class          = NEW_URI(LILV_NS_RDFS "Class");
+	world->uris.rdfs_label          = NEW_URI(LILV_NS_RDFS "label");
+	world->uris.rdfs_seeAlso        = NEW_URI(LILV_NS_RDFS "seeAlso");
+	world->uris.rdfs_subClassOf     = NEW_URI(LILV_NS_RDFS "subClassOf");
+	world->uris.xsd_base64Binary    = NEW_URI(LILV_NS_XSD  "base64Binary");
+	world->uris.xsd_boolean         = NEW_URI(LILV_NS_XSD  "boolean");
+	world->uris.xsd_decimal         = NEW_URI(LILV_NS_XSD  "decimal");
+	world->uris.xsd_double          = NEW_URI(LILV_NS_XSD  "double");
+	world->uris.xsd_integer         = NEW_URI(LILV_NS_XSD  "integer");
+	world->uris.null_uri            = NULL;
 
 	world->lv2_plugin_class = lilv_plugin_class_new(
-		world, NULL, world->lv2_plugin_node, "Plugin");
+		world, NULL, world->uris.lv2_Plugin, "Plugin");
 	assert(world->lv2_plugin_class);
 
 	world->n_read_files        = 0;
@@ -107,33 +113,9 @@ lilv_world_free(LilvWorld* world)
 	lilv_plugin_class_free(world->lv2_plugin_class);
 	world->lv2_plugin_class = NULL;
 
-	sord_node_free(world->world, world->dc_replaces_node);
-	sord_node_free(world->world, world->dyn_manifest_node);
-	sord_node_free(world->world, world->lv2_binary_node);
-	sord_node_free(world->world, world->lv2_default_node);
-	sord_node_free(world->world, world->lv2_index_node);
-	sord_node_free(world->world, world->lv2_maximum_node);
-	sord_node_free(world->world, world->lv2_minimum_node);
-	sord_node_free(world->world, world->lv2_plugin_node);
-	sord_node_free(world->world, world->lv2_port_node);
-	sord_node_free(world->world, world->lv2_portproperty_node);
-	sord_node_free(world->world, world->lv2_reportslatency_node);
-	sord_node_free(world->world, world->lv2_specification_node);
-	sord_node_free(world->world, world->lv2_symbol_node);
-	sord_node_free(world->world, world->rdf_a_node);
-	sord_node_free(world->world, world->rdf_value_node);
-	sord_node_free(world->world, world->rdfs_class_node);
-	sord_node_free(world->world, world->rdfs_label_node);
-	sord_node_free(world->world, world->rdfs_seealso_node);
-	sord_node_free(world->world, world->rdfs_subclassof_node);
-	sord_node_free(world->world, world->xsd_boolean_node);
-	sord_node_free(world->world, world->xsd_decimal_node);
-	sord_node_free(world->world, world->xsd_double_node);
-	sord_node_free(world->world, world->xsd_integer_node);
-	lilv_node_free(world->doap_name_val);
-	lilv_node_free(world->lv2_name_val);
-	lilv_node_free(world->lv2_optionalFeature_val);
-	lilv_node_free(world->lv2_requiredFeature_val);
+	for (SordNode** n = (SordNode**)&world->uris; *n; ++n) {
+		sord_node_free(world->world, *n);
+	}
 
 	for (LilvSpec* spec = world->specs; spec;) {
 		LilvSpec* next = spec->next;
@@ -149,10 +131,18 @@ lilv_world_free(LilvWorld* world)
 		const LilvPlugin* p = lilv_plugins_get(world->plugins, i);
 		lilv_plugin_free((LilvPlugin*)p);
 	}
-	zix_tree_free(world->plugins);
+	zix_tree_free((ZixTree*)world->plugins);
 	world->plugins = NULL;
 
-	zix_tree_free(world->plugin_classes);
+	zix_tree_free((ZixTree*)world->loaded_files);
+	world->loaded_files = NULL;
+
+#ifdef LILV_NEW_LV2
+	zix_tree_free((ZixTree*)world->libs);
+	world->libs = NULL;
+#endif
+
+	zix_tree_free((ZixTree*)world->plugin_classes);
 	world->plugin_classes = NULL;
 
 	sord_free(world->model);
@@ -166,8 +156,8 @@ lilv_world_free(LilvWorld* world)
 
 LILV_API
 void
-lilv_world_set_option(LilvWorld*       world,
-                      const char*      option,
+lilv_world_set_option(LilvWorld*      world,
+                      const char*     option,
                       const LilvNode* value)
 {
 	if (!strcmp(option, LILV_OPTION_DYN_MANIFEST)) {
@@ -184,18 +174,6 @@ lilv_world_set_option(LilvWorld*       world,
 	LILV_WARNF("Unrecognized or invalid option `%s'\n", option);
 }
 
-static SordIter*
-lilv_world_find_statements(const LilvWorld* world,
-                           SordModel*       model,
-                           const SordNode*  subject,
-                           const SordNode*  predicate,
-                           const SordNode*  object,
-                           const SordNode*  graph)
-{
-	SordQuad pat = { subject, predicate, object, graph };
-	return sord_find(model, pat);
-}
-
 LILV_API
 LilvNodes*
 lilv_world_find_nodes(LilvWorld*      world,
@@ -203,7 +181,7 @@ lilv_world_find_nodes(LilvWorld*      world,
                       const LilvNode* predicate,
                       const LilvNode* object)
 {
-	if (!lilv_node_is_uri(subject) && !lilv_node_is_blank(subject)) {
+	if (subject && !lilv_node_is_uri(subject) && !lilv_node_is_blank(subject)) {
 		LILV_ERRORF("Subject `%s' is not a resource\n", subject->str_val);
 		return NULL;
 	}
@@ -211,18 +189,24 @@ lilv_world_find_nodes(LilvWorld*      world,
 		LILV_ERRORF("Predicate `%s' is not a URI\n", predicate->str_val);
 		return NULL;
 	}
+	if (!subject && !object) {
+		LILV_ERROR("Both subject and object are NULL\n");
+		return NULL;
+	}
 
-	SordNode* subject_node = (lilv_node_is_uri(subject))
+	SordNode* const subject_node = subject
 		? sord_node_copy(subject->val.uri_val)
-		: sord_new_blank(world->world,
-		                 (const uint8_t*)lilv_node_as_blank(subject));
+		: NULL;
 
-	LilvNodes* ret = lilv_world_query_values_internal(world,
-	                                                  subject_node,
-	                                                  predicate->val.uri_val,
-	                                                  NULL);
+	SordNode* const object_node = object
+		? sord_node_copy(object->val.uri_val)
+		: NULL;
+
+	LilvNodes* ret = lilv_world_query_values_internal(
+		world, subject_node, predicate->val.uri_val, object_node);
 
 	sord_node_free(world->world, subject_node);
+	sord_node_free(world->world, object_node);
 	return ret;
 }
 
@@ -232,8 +216,7 @@ lilv_world_query_internal(LilvWorld*      world,
                           const SordNode* predicate,
                           const SordNode* object)
 {
-	return lilv_world_find_statements(world, world->model,
-	                                  subject, predicate, object, NULL);
+	return sord_search(world->model, subject, predicate, object, NULL);
 }
 
 LilvNodes*
@@ -244,7 +227,8 @@ lilv_world_query_values_internal(LilvWorld*      world,
 {
 	return lilv_nodes_from_stream_objects(
 		world,
-		lilv_world_query_internal(world, subject, predicate, object));
+		lilv_world_query_internal(world, subject, predicate, object),
+		(object == NULL) ? SORD_OBJECT : SORD_SUBJECT);
 }
 
 static SerdNode
@@ -300,25 +284,25 @@ lilv_world_add_spec(LilvWorld*      world,
                     const SordNode* specification_node,
                     const SordNode* bundle_node)
 {
-	LilvSpec* spec = malloc(sizeof(struct LilvSpecImpl));
+	LilvSpec* spec = (LilvSpec*)malloc(sizeof(LilvSpec));
 	spec->spec      = sord_node_copy(specification_node);
 	spec->bundle    = sord_node_copy(bundle_node);
 	spec->data_uris = lilv_nodes_new();
 
 	// Add all plugin data files (rdfs:seeAlso)
-	SordIter* files = lilv_world_find_statements(
-		world, world->model,
+	SordIter* files = sord_search(
+		world->model,
 		specification_node,
-		world->rdfs_seealso_node,
+		world->uris.rdfs_seeAlso,
 		NULL,
 		NULL);
 	FOREACH_MATCH(files) {
-		const SordNode* file_node = lilv_match_object(files);
-		zix_tree_insert(spec->data_uris,
+		const SordNode* file_node = sord_iter_get_node(files, SORD_OBJECT);
+		zix_tree_insert((ZixTree*)spec->data_uris,
 		                lilv_node_new_from_node(world, file_node),
 		                NULL);
 	}
-	lilv_match_end(files);
+	sord_iter_free(files);
 
 	// Add specification to world specification list
 	spec->next   = world->specs;
@@ -326,15 +310,16 @@ lilv_world_add_spec(LilvWorld*      world,
 }
 
 static void
-lilv_world_add_plugin(LilvWorld*      world,
-                      const SordNode* plugin_node,
-                      SerdNode*       manifest_uri,
-                      const SordNode* dyn_manifest_lib,
-                      const SordNode* bundle_node)
+lilv_world_add_plugin(LilvWorld*       world,
+                      const SordNode*  plugin_node,
+                      SerdNode*        manifest_uri,
+                      void*            dynmanifest,
+                      const SordNode*  bundle_node)
 {
 	LilvNode* plugin_uri = lilv_node_new_from_node(world, plugin_node);
 
-	const LilvPlugin* last = lilv_plugins_get_by_uri(world->plugins, plugin_uri);
+	const LilvPlugin* last = lilv_plugins_get_by_uri(world->plugins,
+	                                                 plugin_uri);
 	if (last) {
 		LILV_ERRORF("Duplicate plugin <%s>\n", lilv_node_as_uri(plugin_uri));
 		LILV_ERRORF("... found in %s\n", lilv_node_as_string(
@@ -349,32 +334,35 @@ lilv_world_add_plugin(LilvWorld*      world,
 	LilvPlugin* plugin     = lilv_plugin_new(world, plugin_uri, bundle_uri);
 
 	// Add manifest as plugin data file (as if it were rdfs:seeAlso)
-	zix_tree_insert(plugin->data_uris,
+	zix_tree_insert((ZixTree*)plugin->data_uris,
 	                lilv_new_uri(world, (const char*)manifest_uri->buf),
 	                NULL);
 
+#ifdef LILV_DYN_MANIFEST
 	// Set dynamic manifest library URI, if applicable
-	if (dyn_manifest_lib) {
-		plugin->dynman_uri = lilv_node_new_from_node(world, dyn_manifest_lib);
+	if (dynmanifest) {
+		plugin->dynmanifest = (LilvDynManifest*)dynmanifest;
+		++((LilvDynManifest*)dynmanifest)->refs;
 	}
+#endif
 
 	// Add all plugin data files (rdfs:seeAlso)
-	SordIter* files = lilv_world_find_statements(
-		world, world->model,
+	SordIter* files = sord_search(
+		world->model,
 		plugin_node,
-		world->rdfs_seealso_node,
+		world->uris.rdfs_seeAlso,
 		NULL,
 		NULL);
 	FOREACH_MATCH(files) {
-		const SordNode* file_node = lilv_match_object(files);
-		zix_tree_insert(plugin->data_uris,
+		const SordNode* file_node = sord_iter_get_node(files, SORD_OBJECT);
+		zix_tree_insert((ZixTree*)plugin->data_uris,
 		                lilv_node_new_from_node(world, file_node),
 		                NULL);
 	}
-	lilv_match_end(files);
+	sord_iter_free(files);
 
 	// Add plugin to world plugin sequence
-	zix_tree_insert(world->plugins, plugin, NULL);
+	zix_tree_insert((ZixTree*)world->plugins, plugin, NULL);
 }
 
 static void
@@ -391,53 +379,54 @@ lilv_world_load_dyn_manifest(LilvWorld* world,
 	LV2_Dyn_Manifest_Handle handle = NULL;
 
 	// ?dman a dynman:DynManifest
-	SordIter* dmanifests = lilv_world_find_statements(
-		world, world->model,
+	SordIter* dmanifests = sord_search(
+		world->model,
 		NULL,
-		world->rdf_a_node,
-		world->dyn_manifest_node,
+		world->uris.rdf_a,
+		world->uris.dman_DynManifest,
 		bundle_node);
 	FOREACH_MATCH(dmanifests) {
-		const SordNode* dmanifest = lilv_match_subject(dmanifests);
+		const SordNode* dmanifest = sord_iter_get_node(dmanifests, SORD_SUBJECT);
 
 		// ?dman lv2:binary ?binary
-		SordIter* binaries  = lilv_world_find_statements(
-			world, world->model,
+		SordIter* binaries = sord_search(
+			world->model,
 			dmanifest,
-			world->lv2_binary_node,
+			world->uris.lv2_binary,
 			NULL,
 			bundle_node);
-		if (lilv_matches_end(binaries)) {
-			lilv_match_end(binaries);
+		if (sord_iter_end(binaries)) {
+			sord_iter_free(binaries);
 			LILV_ERRORF("Dynamic manifest in <%s> has no binaries, ignored\n",
 			            sord_node_get_string(bundle_node));
 			continue;
 		}
 
 		// Get binary path
-		const SordNode* binary   = lilv_match_object(binaries);
+		const SordNode* binary   = sord_iter_get_node(binaries, SORD_OBJECT);
 		const uint8_t*  lib_uri  = sord_node_get_string(binary);
 		const char*     lib_path = lilv_uri_to_path((const char*)lib_uri);
 		if (!lib_path) {
 			LILV_ERROR("No dynamic manifest library path\n");
-			lilv_match_end(binaries);
+			sord_iter_free(binaries);
 			continue;
 		}
 
 		// Open library
 		void* lib = dlopen(lib_path, RTLD_LAZY);
 		if (!lib) {
-			LILV_ERRORF("Failed to open dynamic manifest library `%s'\n", lib_path);
-			lilv_match_end(binaries);
+			LILV_ERRORF("Failed to open dynmanifest library `%s'\n", lib_path);
+			sord_iter_free(binaries);
 			continue;
 		}
 
 		// Open dynamic manifest
-		typedef int (*OpenFunc)(LV2_Dyn_Manifest_Handle*, const LV2_Feature *const *);
-		OpenFunc open_func = (OpenFunc)lilv_dlfunc(lib, "lv2_dyn_manifest_open");
-		if (!open_func || open_func(&handle, &dman_features)) {
-			LILV_ERRORF("Failed to find `lv2_dyn_manifest_open' in `%s'\n", lib_path);
-			lilv_match_end(binaries);
+		typedef int (*OpenFunc)(LV2_Dyn_Manifest_Handle*,
+		                        const LV2_Feature *const *);
+		OpenFunc dmopen = (OpenFunc)lilv_dlfunc(lib, "lv2_dyn_manifest_open");
+		if (!dmopen || dmopen(&handle, &dman_features)) {
+			LILV_ERRORF("No `lv2_dyn_manifest_open' in `%s'\n", lib_path);
+			sord_iter_free(binaries);
 			dlclose(lib);
 			continue;
 		}
@@ -447,22 +436,31 @@ lilv_world_load_dyn_manifest(LilvWorld* world,
 		GetSubjectsFunc get_subjects_func = (GetSubjectsFunc)lilv_dlfunc(
 			lib, "lv2_dyn_manifest_get_subjects");
 		if (!get_subjects_func) {
-			LILV_ERRORF("Failed to find `lv2_dyn_manifest_get_subjects' in `%s'\n",
+			LILV_ERRORF("No `lv2_dyn_manifest_get_subjects' in `%s'\n",
 			            lib_path);
-			lilv_match_end(binaries);
+			sord_iter_free(binaries);
 			dlclose(lib);
 			continue;
 		}
 
+		LilvDynManifest* desc = malloc(sizeof(LilvDynManifest));
+		desc->bundle = lilv_node_new_from_node(world, bundle_node);
+		desc->lib    = lib;
+		desc->handle = handle;
+		desc->refs   = 0;
+
 		// Generate data file
 		FILE* fd = tmpfile();
 		get_subjects_func(handle, fd);
 		rewind(fd);
 
 		// Parse generated data file
-		SerdEnv*    env    = serd_env_new(sord_node_to_serd_node(bundle_node));
-		SerdReader* reader = sord_new_reader(
-			world->model, env, SERD_TURTLE, bundle_node);
+		const SerdNode* base   = sord_node_to_serd_node(dmanifest);
+		SerdEnv*        env    = serd_env_new(base);
+		SerdReader*     reader = sord_new_reader(
+			world->model, env, SERD_TURTLE, sord_node_copy(dmanifest));
+		serd_reader_add_blank_prefix(reader,
+		                             lilv_world_blank_node_prefix(world));
 		serd_reader_read_file_handle(reader, fd,
 		                             (const uint8_t*)"(dyn-manifest)");
 		serd_reader_free(reader);
@@ -472,23 +470,22 @@ lilv_world_load_dyn_manifest(LilvWorld* world,
 		fclose(fd);
 
 		// ?plugin a lv2:Plugin
-		SordIter* plug_results = lilv_world_find_statements(
-			world, world->model,
+		SordIter* plug_results = sord_search(
+			world->model,
 			NULL,
-			world->rdf_a_node,
-			world->lv2_plugin_node,
-			bundle_node);
+			world->uris.rdf_a,
+			world->uris.lv2_Plugin,
+			dmanifest);
 		FOREACH_MATCH(plug_results) {
-			const SordNode* plugin_node = lilv_match_subject(plug_results);
+			const SordNode* plugin_node = sord_iter_get_node(plug_results, SORD_SUBJECT);
 			lilv_world_add_plugin(world, plugin_node,
-			                      &manifest_uri, binary, bundle_node);
+			                      &manifest_uri, desc, bundle_node);
 		}
-		lilv_match_end(plug_results);
+		sord_iter_free(plug_results);
 
-		lilv_match_end(binaries);
-		dlclose(lib);
+		sord_iter_free(binaries);
 	}
-	lilv_match_end(dmanifests);
+	sord_iter_free(dmanifests);
 #endif  // LILV_DYN_MANIFEST
 }
 
@@ -521,108 +518,65 @@ lilv_world_load_bundle(LilvWorld* world, LilvNode* bundle_uri)
 	}
 
 	// ?plugin a lv2:Plugin
-	SordIter* plug_results = lilv_world_find_statements(
-		world, world->model,
+	SordIter* plug_results = sord_search(
+		world->model,
 		NULL,
-		world->rdf_a_node,
-		world->lv2_plugin_node,
+		world->uris.rdf_a,
+		world->uris.lv2_Plugin,
 		bundle_node);
 	FOREACH_MATCH(plug_results) {
-		const SordNode* plugin_node = lilv_match_subject(plug_results);
+		const SordNode* plugin_node = sord_iter_get_node(plug_results, SORD_SUBJECT);
 		lilv_world_add_plugin(world, plugin_node,
 		                      &manifest_uri, NULL, bundle_node);
 	}
-	lilv_match_end(plug_results);
+	sord_iter_free(plug_results);
 
 	lilv_world_load_dyn_manifest(world, bundle_node, manifest_uri);
 
 	// ?specification a lv2:Specification
-	SordIter* spec_results = lilv_world_find_statements(
-		world, world->model,
+	SordIter* spec_results = sord_search(
+		world->model,
 		NULL,
-		world->rdf_a_node,
-		world->lv2_specification_node,
+		world->uris.rdf_a,
+		world->uris.lv2_Specification,
 		bundle_node);
 	FOREACH_MATCH(spec_results) {
-		const SordNode* spec = lilv_match_subject(spec_results);
+		const SordNode* spec = sord_iter_get_node(spec_results, SORD_SUBJECT);
 		lilv_world_add_spec(world, spec, bundle_node);
 	}
-	lilv_match_end(spec_results);
+	sord_iter_free(spec_results);
 
 	serd_node_free(&manifest_uri);
 }
 
-/** Expand variables (e.g. POSIX ~ or $FOO, Windows %FOO%) in @a path. */
-static char*
-expand(const char* path)
+static void
+load_dir_entry(const char* dir, const char* name, void* data)
 {
-#ifdef HAVE_WORDEXP
-	char*     ret = NULL;
-	wordexp_t p;
-	wordexp(path, &p, 0);
-	if (p.we_wordc == 0) {
-		/* Literal directory path (e.g. no variables or ~) */
-		ret = lilv_strdup(path);
-	} else if (p.we_wordc == 1) {
-		/* Directory path expands (e.g. contains ~ or $FOO) */
-		ret = lilv_strdup(p.we_wordv[0]);
-	} else {
-		/* Multiple expansions in a single directory path? */
-		LILV_ERRORF("Malformed path `%s' ignored\n", path);
-	}
-	wordfree(&p);
-#elif defined(__WIN32__)
-	static const size_t len = 32767;
-	char*               ret = malloc(len);
-	ExpandEnvironmentStrings(path, ret, len);
-#else
-	char* ret = lilv_strdup(path);
-#endif
-	return ret;
+	LilvWorld* world = (LilvWorld*)data;
+	if (!strcmp(name, ".") || !strcmp(name, ".."))
+		return;
+
+	const char* scheme  = (dir[0] == '/') ? "file://" : "file:///";
+	char*       uri     = lilv_strjoin(scheme, dir, "/", name, "/", NULL);
+	LilvNode*   uri_val = lilv_new_uri(world, uri);
+
+	lilv_world_load_bundle(world, uri_val);
+	lilv_node_free(uri_val);
+	free(uri);
 }
 
 /** Load all bundles in the directory at @a dir_path. */
 static void
 lilv_world_load_directory(LilvWorld* world, const char* dir_path)
 {
-	char* path = expand(dir_path);
+	char* path = lilv_expand(dir_path);
 	if (!path) {
 		LILV_WARNF("Empty path `%s'\n", path);
 		return;
 	}
 
-	DIR* pdir = opendir(path);
-	if (!pdir) {
-		free(path);
-		return;
-	}
-
-#ifdef __WIN32__
-	static const char* const file_scheme = "file:///";
-#else
-	static const char* const file_scheme = "file://";
-#endif
-
-	struct dirent  entry;
-	struct dirent* result;
-	while (!readdir_r(pdir, &entry, &result) && result) {
-		if (!strcmp(entry.d_name, ".") || !strcmp(entry.d_name, ".."))
-			continue;
-
-		char* uri = lilv_strjoin(file_scheme,
-		                         path, "/",
-		                         entry.d_name, "/",
-		                         NULL);
-
-		LilvNode* uri_val = lilv_new_uri(world, uri);
-		lilv_world_load_bundle(world, uri_val);
-		lilv_node_free(uri_val);
-
-		free(uri);
-	}
-
+	lilv_dir_for_each(path, world, load_dir_entry);
 	free(path);
-	closedir(pdir);
 }
 
 static bool
@@ -655,7 +609,7 @@ lilv_world_load_path(LilvWorld*  world,
 		const char* const sep = first_path_sep(lv2_path);
 		if (sep) {
 			const size_t dir_len = sep - lv2_path;
-			char* const  dir     = malloc(dir_len + 1);
+			char* const  dir     = (char*)malloc(dir_len + 1);
 			memcpy(dir, lv2_path, dir_len);
 			dir[dir_len] = '\0';
 			lilv_world_load_directory(world, dir);
@@ -673,15 +627,15 @@ lilv_world_load_specifications(LilvWorld* world)
 {
 	for (LilvSpec* spec = world->specs; spec; spec = spec->next) {
 		LILV_FOREACH(nodes, f, spec->data_uris) {
-			LilvNode*      file     = lilv_collection_get(spec->data_uris, f);
-			const uint8_t* file_uri = (const uint8_t*)lilv_node_as_uri(file);
-			SerdEnv*       env      = serd_env_new(
-				sord_node_to_serd_node(file->val.uri_val));
-			SerdReader*    reader   = sord_new_reader(world->model, env,
-			                                          SERD_TURTLE, NULL);
+			LilvNode* file = (LilvNode*)lilv_collection_get(spec->data_uris, f);
+
+			const SerdNode* node   = sord_node_to_serd_node(file->val.uri_val);
+			SerdEnv*        env    = serd_env_new(node);
+			SerdReader*     reader = sord_new_reader(world->model, env,
+			                                         SERD_TURTLE, NULL);
 			serd_reader_add_blank_prefix(reader,
 			                             lilv_world_blank_node_prefix(world));
-			serd_reader_read_file(reader, file_uri);
+			serd_reader_read_file(reader, node->buf);
 			serd_reader_free(reader);
 			serd_env_free(env);
 		}
@@ -697,30 +651,30 @@ lilv_world_load_plugin_classes(LilvWorld* world)
 	   a menu), they won't be seen anyway...
 	*/
 
-	SordIter* classes = lilv_world_find_statements(
-		world, world->model,
+	SordIter* classes = sord_search(
+		world->model,
 		NULL,
-		world->rdf_a_node,
-		world->rdfs_class_node,
+		world->uris.rdf_a,
+		world->uris.rdfs_Class,
 		NULL);
 	FOREACH_MATCH(classes) {
-		const SordNode* class_node = lilv_match_subject(classes);
+		const SordNode* class_node = sord_iter_get_node(classes, SORD_SUBJECT);
 
 		// Get parents (superclasses)
-		SordIter* parents = lilv_world_find_statements(
-			world, world->model,
+		SordIter* parents = sord_search(
+			world->model,
 			class_node,
-			world->rdfs_subclassof_node,
+			world->uris.rdfs_subClassOf,
 			NULL,
 			NULL);
 
-		if (lilv_matches_end(parents)) {
-			lilv_match_end(parents);
+		if (sord_iter_end(parents)) {
+			sord_iter_free(parents);
 			continue;
 		}
 
-		const SordNode* parent_node = lilv_match_object(parents);
-		lilv_match_end(parents);
+		const SordNode* parent_node = sord_iter_get_node(parents, SORD_OBJECT);
+		sord_iter_free(parents);
 
 		if (!sord_node_get_type(parent_node) == SORD_URI) {
 			// Class parent is not a resource, ignore (e.g. owl restriction)
@@ -728,31 +682,31 @@ lilv_world_load_plugin_classes(LilvWorld* world)
 		}
 
 		// Get labels
-		SordIter* labels = lilv_world_find_statements(
-			world, world->model,
+		SordIter* labels = sord_search(
+			world->model,
 			class_node,
-			world->rdfs_label_node,
+			world->uris.rdfs_label,
 			NULL,
 			NULL);
 
-		if (lilv_matches_end(labels)) {
-			lilv_match_end(labels);
+		if (sord_iter_end(labels)) {
+			sord_iter_free(labels);
 			continue;
 		}
 
-		const SordNode* label_node = lilv_match_object(labels);
+		const SordNode* label_node = sord_iter_get_node(labels, SORD_OBJECT);
 		const uint8_t*  label      = sord_node_get_string(label_node);
-		lilv_match_end(labels);
+		sord_iter_free(labels);
 
 		LilvPluginClasses* classes = world->plugin_classes;
 		LilvPluginClass*   pclass  = lilv_plugin_class_new(
 			world, parent_node, class_node, (const char*)label);
 
 		if (pclass) {
-			zix_tree_insert(classes, pclass, NULL);
+			zix_tree_insert((ZixTree*)classes, pclass, NULL);
 		}
 	}
-	lilv_match_end(classes);
+	sord_iter_free(classes);
 }
 
 LILV_API
@@ -767,15 +721,15 @@ lilv_world_load_all(LilvWorld* world)
 	lilv_world_load_path(world, lv2_path);
 
 	LILV_FOREACH(plugins, p, world->plugins) {
-		const LilvPlugin* plugin     = lilv_collection_get(world->plugins, p);
-		const LilvNode*   plugin_uri = lilv_plugin_get_uri(plugin);
+		const LilvPlugin* plugin = (const LilvPlugin*)lilv_collection_get(
+			(ZixTree*)world->plugins, p);
 
 		// ?new dc:replaces plugin
-		SordIter* replacement = lilv_world_find_statements(
-			world, world->model,
+		SordIter* replacement = sord_search(
+			world->model,
 			NULL,
-			world->dc_replaces_node,
-			lilv_node_as_node(plugin_uri),
+			world->uris.dc_replaces,
+			lilv_node_as_node(lilv_plugin_get_uri(plugin)),
 			NULL);
 		if (!sord_iter_end(replacement)) {
 			/* TODO: Check if replacement is actually a known plugin,
@@ -783,7 +737,7 @@ lilv_world_load_all(LilvWorld* world)
 			*/
 			((LilvPlugin*)plugin)->replaced = true;
 		}
-		lilv_match_end(replacement);
+		sord_iter_free(replacement);
 	}
 
 	// Query out things to cache
@@ -791,6 +745,55 @@ lilv_world_load_all(LilvWorld* world)
 	lilv_world_load_plugin_classes(world);
 }
 
+LILV_API
+int
+lilv_world_load_resource(LilvWorld*      world,
+                         const LilvNode* resource)
+{
+	if (!lilv_node_is_uri(resource) && !lilv_node_is_blank(resource)) {
+		LILV_ERRORF("Node `%s' is not a resource\n", resource->str_val);
+		return -1;
+	}
+
+	int       n_read = 0;
+	SordIter* files  = sord_search(world->model,
+	                               resource->val.uri_val,
+	                               world->uris.rdfs_seeAlso,
+	                               NULL, NULL);
+	FOREACH_MATCH(files) {
+		const SordNode* file      = sord_iter_get_node(files, SORD_OBJECT);
+		const uint8_t*  str       = sord_node_get_string(file);
+		LilvNode*       file_node = lilv_node_new_from_node(world, file);
+		ZixTreeIter*    iter;
+		if (zix_tree_find((ZixTree*)world->loaded_files, file_node, &iter)) {
+			if (sord_node_get_type(file) == SORD_URI) {
+				const SerdNode* base   = sord_node_to_serd_node(file);
+				SerdEnv*        env    = serd_env_new(base);
+				SerdReader*     reader = sord_new_reader(
+					world->model, env, SERD_TURTLE, (SordNode*)file);
+				serd_reader_add_blank_prefix(
+					reader, lilv_world_blank_node_prefix(world));
+				if (!serd_reader_read_file(reader, str)) {
+					++n_read;
+					zix_tree_insert(
+						(ZixTree*)world->loaded_files, file_node, NULL);
+					file_node = NULL;  // prevent deletion...
+				} else {
+					LILV_ERRORF("Error loading resource `%s'\n", str);
+				}
+				serd_reader_free(reader);
+				serd_env_free(env);
+			} else {
+				LILV_ERRORF("rdfs:seeAlso node `%s' is not a URI\n", str);
+			}
+		}
+		lilv_node_free(file_node);  // ...here
+	}
+	sord_iter_free(files);
+
+	return n_read;
+}
+
 LILV_API
 const LilvPluginClass*
 lilv_world_get_plugin_class(const LilvWorld* world)
