@@ -15,22 +15,45 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#define _XOPEN_SOURCE 600
+#define _POSIX_C_SOURCE 200112L /* for setenv */
 
-#include <unistd.h>
-#include <string.h>
+#include <assert.h>
+#include <ctype.h>
+#include <errno.h>
+#include <float.h>
+#include <limits.h>
+#include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <assert.h>
+#include <string.h>
 #include <sys/stat.h>
-#include <limits.h>
-#include <float.h>
-#include <math.h>
+
+#ifdef _WIN32
+#    include <direct.h>
+#    define mkdir(path, flags) _mkdir(path)
+#    define setenv(n, v, r) SetEnvironmentVariable((n), (v))
+#    define unsetenv(n) SetEnvironmentVariable((n), NULL)
+#else
+#    include <dirent.h>
+#    include <unistd.h>
+#endif
 
 #include "lilv/lilv.h"
+#include "../src/lilv_internal.h"
+
+#include "lv2/lv2plug.in/ns/ext/state/state.h"
+#include "lv2/lv2plug.in/ns/ext/urid/urid.h"
 
 #define TEST_PATH_MAX 1024
 
+#if defined(__APPLE__)
+#    define SHLIB_EXT ".dylib"
+#elif defined(_WIN32)
+#    define SHLIB_EXT ".dll"
+#else
+#    define SHLIB_EXT ".so"
+#endif
+
 static char bundle_dir_name[TEST_PATH_MAX];
 static char bundle_dir_uri[TEST_PATH_MAX];
 static char manifest_name[TEST_PATH_MAX];
@@ -46,7 +69,7 @@ delete_bundle(void)
 {
 	unlink(content_name);
 	unlink(manifest_name);
-	rmdir(bundle_dir_name);
+	remove(bundle_dir_name);
 }
 
 void
@@ -102,7 +125,7 @@ load_all_bundles(void)
 void
 create_bundle(char *manifest, char *content)
 {
-	if (mkdir(bundle_dir_name, 0700))
+	if (mkdir(bundle_dir_name, 0700) && errno != EEXIST)
 		fatal_error("Cannot create directory %s\n", bundle_dir_name);
 	write_file(manifest_name, manifest);
 	write_file(content_name, content);
@@ -136,7 +159,7 @@ cleanup(void)
 	test_count++;\
 	if (!(check)) {\
 		error_count++;\
-		fprintf(stderr, "Failure at lilv_test.c:%d: %s\n", __LINE__, #check);\
+		fprintf(stderr, "lilv_test.c:%d: error: %s\n", __LINE__, #check);\
 	}\
 } while (0)
 
@@ -202,7 +225,7 @@ int
 test_value(void)
 {
 	if (!start_bundle(MANIFEST_PREFIXES
-			":plug a lv2:Plugin ; lv2:binary <foo.so> ; rdfs:seeAlso <plugin.ttl> .\n",
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
 			BUNDLE_PREFIXES
 			":plug a lv2:Plugin ; a lv2:CompressorPlugin ; "
 			PLUGIN_NAME("Test plugin") " ; "
@@ -321,7 +344,7 @@ discovery_verify_plugin(const LilvPlugin* plugin)
 		TEST_ASSERT(lib_uri);
 		TEST_ASSERT(lilv_node_is_uri(lib_uri));
 		TEST_ASSERT(lilv_node_as_uri(lib_uri));
-		TEST_ASSERT(strstr(lilv_node_as_uri(lib_uri), "foo.so"));
+		TEST_ASSERT(strstr(lilv_node_as_uri(lib_uri), "foo" SHLIB_EXT));
 		TEST_ASSERT(lilv_plugin_verify(plugin));
 	}
 }
@@ -330,7 +353,7 @@ int
 test_discovery(void)
 {
 	if (!start_bundle(MANIFEST_PREFIXES
-			":plug a lv2:Plugin ; lv2:binary <foo.so> ; rdfs:seeAlso <plugin.ttl> .\n",
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
 			BUNDLE_PREFIXES
 			":plug a lv2:Plugin ;"
 			PLUGIN_NAME("Test plugin") " ; "
@@ -344,9 +367,9 @@ test_discovery(void)
 	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
 	TEST_ASSERT(lilv_plugins_size(plugins) > 0);
 
-	const const LilvPlugin* explug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+	const LilvPlugin* explug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
 	TEST_ASSERT(explug != NULL);
-	const const LilvPlugin* explug2 = lilv_plugins_get_by_uri(plugins, plugin2_uri_value);
+	const LilvPlugin* explug2 = lilv_plugins_get_by_uri(plugins, plugin2_uri_value);
 	TEST_ASSERT(explug2 == NULL);
 
 	if (explug) {
@@ -369,11 +392,46 @@ test_discovery(void)
 
 /*****************************************************************************/
 
+int
+test_lv2_path(void)
+{
+#ifndef _WIN32
+	char* orig_lv2_path = lilv_strdup(getenv("LV2_PATH"));
+
+	setenv("LV2_PATH", "~/.lv2:/usr/local/lib/lv2:/usr/lib/lv2", 1);
+
+	LilvWorld* world = lilv_world_new();
+	lilv_world_load_all(world);
+
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const size_t n_plugins     = lilv_plugins_size(plugins);
+
+	lilv_world_free(world);
+
+	setenv("LV2_PATH", "$HOME/.lv2:/usr/local/lib/lv2:/usr/lib/lv2", 1);
+	world = lilv_world_new();
+	lilv_world_load_all(world);
+	plugins = lilv_world_get_all_plugins(world);
+	TEST_ASSERT(lilv_plugins_size(plugins) == n_plugins);
+	lilv_world_free(world);
+
+	if (orig_lv2_path) {
+		setenv("LV2_PATH", orig_lv2_path, 1);
+	} else {
+		unsetenv("LV2_PATH");
+	}
+	free(orig_lv2_path);
+#endif
+	return 1;
+}
+
+/*****************************************************************************/
+
 int
 test_verify(void)
 {
 	if (!start_bundle(MANIFEST_PREFIXES
-			":plug a lv2:Plugin ; lv2:binary <foo.so> ; rdfs:seeAlso <plugin.ttl> .\n",
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
 			BUNDLE_PREFIXES
 			":plug a lv2:Plugin ; "
 			PLUGIN_NAME("Test plugin") " ; "
@@ -384,7 +442,7 @@ test_verify(void)
 
 	init_uris();
 	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
-	const const LilvPlugin* explug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+	const LilvPlugin* explug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
 	TEST_ASSERT(explug);
 	TEST_ASSERT(lilv_plugin_verify(explug));
 	cleanup_uris();
@@ -397,7 +455,7 @@ int
 test_no_verify(void)
 {
 	if (!start_bundle(MANIFEST_PREFIXES
-			":plug a lv2:Plugin ; lv2:binary <foo.so> ; rdfs:seeAlso <plugin.ttl> .\n",
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
 			BUNDLE_PREFIXES
 			":plug a lv2:Plugin . "))
 		return 0;
@@ -417,7 +475,7 @@ int
 test_classes(void)
 {
 	if (!start_bundle(MANIFEST_PREFIXES
-			":plug a lv2:Plugin ; lv2:binary <foo.so> ; rdfs:seeAlso <plugin.ttl> .\n",
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
 			BUNDLE_PREFIXES
 			":plug a lv2:Plugin ; a lv2:CompressorPlugin ; "
 			PLUGIN_NAME("Test plugin") " ; "
@@ -437,7 +495,7 @@ test_classes(void)
 	TEST_ASSERT(lilv_plugin_classes_size(classes) > lilv_plugin_classes_size(children));
 	TEST_ASSERT(!strcmp(lilv_node_as_string(lilv_plugin_class_get_label(plugin)), "Plugin"));
 	TEST_ASSERT(!strcmp(lilv_node_as_string(lilv_plugin_class_get_uri(plugin)),
-			"http://lv2plug.in/ns/lv2core#Plugin"));
+	                    "http://lv2plug.in/ns/lv2core#Plugin"));
 
 	LILV_FOREACH(plugin_classes, i, children) {
 		TEST_ASSERT(lilv_node_equals(
@@ -461,13 +519,14 @@ int
 test_plugin(void)
 {
 	if (!start_bundle(MANIFEST_PREFIXES
-			":plug a lv2:Plugin ; lv2:binary <foo.so> ; rdfs:seeAlso <plugin.ttl> .\n",
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
 			BUNDLE_PREFIXES
 			":plug a lv2:Plugin ; a lv2:CompressorPlugin ; "
 			PLUGIN_NAME("Test plugin") " ; "
 			LICENSE_GPL " ; "
 			"lv2:optionalFeature lv2:hardRTCapable ; "
 			"lv2:requiredFeature <http://lv2plug.in/ns/ext/event> ; "
+			"lv2:extensionData <http://example.org/extdata> ;"
 			":foo 1.6180 ; "
 			":bar true ; "
 			":baz false ; "
@@ -485,7 +544,8 @@ test_plugin(void)
 			"] , [ "
 			"  a lv2:ControlPort ; a lv2:OutputPort ; "
 			"  lv2:index 2 ; lv2:symbol \"latency\" ; lv2:name \"Latency\" ; "
-			"  lv2:portProperty lv2:reportsLatency "
+			"  lv2:portProperty lv2:reportsLatency ; "
+			"  lv2:designation lv2:latency "
 			"] . \n"
 			":thing doap:name \"Something else\" .\n"))
 		return 0;
@@ -495,9 +555,9 @@ test_plugin(void)
 	const LilvPlugin* plug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
 	TEST_ASSERT(plug);
 
-	const LilvPluginClass* class = lilv_plugin_get_class(plug);
-	const LilvNode* class_uri = lilv_plugin_class_get_uri(class);
-	TEST_ASSERT(!strcmp(lilv_node_as_string(class_uri),
+	const LilvPluginClass* klass = lilv_plugin_get_class(plug);
+	const LilvNode* klass_uri = lilv_plugin_class_get_uri(klass);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(klass_uri),
 			"http://lv2plug.in/ns/lv2core#CompressorPlugin"));
 
 	const LilvNode* plug_bundle_uri = lilv_plugin_get_bundle_uri(plug);
@@ -524,9 +584,9 @@ test_plugin(void)
 	free(manifest_uri);
 	free(data_uri);
 
-	float mins[1];
-	float maxs[1];
-	float defs[1];
+	float mins[3];
+	float maxs[3];
+	float defs[3];
 	lilv_plugin_get_port_ranges_float(plug, mins, maxs, defs);
 	TEST_ASSERT(mins[0] == -1.0f);
 	TEST_ASSERT(maxs[0] == 1.0f);
@@ -553,6 +613,15 @@ test_plugin(void)
 	TEST_ASSERT(lilv_plugin_has_latency(plug));
 	TEST_ASSERT(lilv_plugin_get_latency_port_index(plug) == 2);
 
+	LilvNode* lv2_latency = lilv_new_uri(world,
+			"http://lv2plug.in/ns/lv2core#latency");
+	LilvPort* latency_port = lilv_plugin_get_port_by_designation(
+		plug, out_class, lv2_latency);
+	lilv_node_free(lv2_latency);
+
+	TEST_ASSERT(latency_port);
+	TEST_ASSERT(lilv_port_get_index(plug, latency_port) == 2);
+
 	LilvNode* rt_feature = lilv_new_uri(world,
 			"http://lv2plug.in/ns/lv2core#hardRTCapable");
 	LilvNode* event_feature = lilv_new_uri(world,
@@ -637,6 +706,17 @@ test_plugin(void)
 	TEST_ASSERT(lilv_uis_size(uis) == 0);
 	lilv_uis_free(uis);
 
+	LilvNode* extdata = lilv_new_uri(world, "http://example.org/extdata");
+	LilvNode* noextdata = lilv_new_uri(world, "http://example.org/noextdata");
+	LilvNodes* extdatas = lilv_plugin_get_extension_data(plug);
+	TEST_ASSERT(lilv_plugin_has_extension_data(plug, extdata));
+	TEST_ASSERT(!lilv_plugin_has_extension_data(plug, noextdata));
+	TEST_ASSERT(lilv_nodes_size(extdatas) == 1);
+	TEST_ASSERT(lilv_node_equals(lilv_nodes_get_first(extdatas), extdata));
+	lilv_node_free(noextdata);
+	lilv_node_free(extdata);
+	lilv_nodes_free(extdatas);
+
 	lilv_nodes_free(thing_names);
 	lilv_node_free(thing_uri);
 	lilv_node_free(name_p);
@@ -654,7 +734,7 @@ int
 test_port(void)
 {
 	if (!start_bundle(MANIFEST_PREFIXES
-			":plug a lv2:Plugin ; lv2:binary <foo.so> ; rdfs:seeAlso <plugin.ttl> .\n",
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
 			BUNDLE_PREFIXES PREFIX_LV2EV
 			":plug a lv2:Plugin ; "
 			PLUGIN_NAME("Test plugin") " ; "
@@ -666,6 +746,7 @@ test_port(void)
 			"  lv2:name \"store\" ; "
 			"  lv2:name \"dépanneur\"@fr-ca ; lv2:name \"épicerie\"@fr-fr ; "
 			"  lv2:name \"tienda\"@es ; "
+			"  rdfs:comment \"comment\"@en , \"commentaires\"@fr ; "
      		"  lv2:portProperty lv2:integer ; "
 			"  lv2:minimum -1.0 ; lv2:maximum 1.0 ; lv2:default 0.5 ; "
 			"  lv2:scalePoint [ rdfs:label \"Sin\"; rdf:value 3 ] ; "
@@ -680,18 +761,18 @@ test_port(void)
 
 	init_uris();
 	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
-	const LilvPlugin* plug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+	const LilvPlugin*  plug    = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
 	TEST_ASSERT(plug);
 
-	LilvNode* psym = lilv_new_string(world, "foo");
-	const LilvPort*  p    = lilv_plugin_get_port_by_index(plug, 0);
-	const LilvPort*  p2   = lilv_plugin_get_port_by_symbol(plug, psym);
+	LilvNode*       psym = lilv_new_string(world, "foo");
+	const LilvPort* p    = lilv_plugin_get_port_by_index(plug, 0);
+	const LilvPort* p2   = lilv_plugin_get_port_by_symbol(plug, psym);
 	lilv_node_free(psym);
 	TEST_ASSERT(p != NULL);
 	TEST_ASSERT(p2 != NULL);
 	TEST_ASSERT(p == p2);
 
-	LilvNode*      nopsym = lilv_new_string(world, "thisaintnoportfoo");
+	LilvNode*       nopsym = lilv_new_string(world, "thisaintnoportfoo");
 	const LilvPort* p3     = lilv_plugin_get_port_by_symbol(plug, nopsym);
 	TEST_ASSERT(p3 == NULL);
 	lilv_node_free(nopsym);
@@ -744,6 +825,36 @@ test_port(void)
 	TEST_ASSERT(!strcmp(lilv_node_as_string(name), "tienda"));
 	lilv_node_free(name);
 
+	// No language match (choose untranslated value)
+	setenv("LANG", "cn", 1);
+	name = lilv_port_get_name(plug, p);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(name), "store"));
+	lilv_node_free(name);
+
+	setenv("LANG", "en_CA.utf-8", 1);
+
+	// Language tagged value with no untranslated values
+	LilvNode*  rdfs_comment = lilv_new_uri(world, LILV_NS_RDFS "comment");
+	LilvNodes* comments     = lilv_port_get_value(plug, p, rdfs_comment);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(lilv_nodes_get_first(comments)),
+	                    "comment"));
+	lilv_nodes_free(comments);
+
+	setenv("LANG", "fr", 1);
+
+	comments = lilv_port_get_value(plug, p, rdfs_comment);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(lilv_nodes_get_first(comments)),
+	                    "commentaires"));
+	lilv_nodes_free(comments);
+
+	setenv("LANG", "cn", 1);
+
+	comments = lilv_port_get_value(plug, p, rdfs_comment);
+	TEST_ASSERT(!comments);
+	lilv_nodes_free(comments);
+
+	lilv_node_free(rdfs_comment);
+
 	setenv("LANG", "C", 1);  // Reset locale
 
 	LilvScalePoints* points = lilv_port_get_scale_points(plug, p);
@@ -756,10 +867,18 @@ test_port(void)
 	const LilvScalePoint* sp1 = lilv_scale_points_get(points, sp_iter);
 	TEST_ASSERT(sp1);
 
-	TEST_ASSERT(!strcmp(lilv_node_as_string(lilv_scale_point_get_label(sp0)), "Sin"));
-	TEST_ASSERT(lilv_node_as_float(lilv_scale_point_get_value(sp0)) == 3);
-	TEST_ASSERT(!strcmp(lilv_node_as_string(lilv_scale_point_get_label(sp1)), "Cos"));
-	TEST_ASSERT(lilv_node_as_float(lilv_scale_point_get_value(sp1)) == 4);
+	TEST_ASSERT(
+		((!strcmp(lilv_node_as_string(lilv_scale_point_get_label(sp0)), "Sin")
+		  && lilv_node_as_float(lilv_scale_point_get_value(sp0)) == 3)
+		 &&
+		 (!strcmp(lilv_node_as_string(lilv_scale_point_get_label(sp1)), "Cos")
+		  && lilv_node_as_float(lilv_scale_point_get_value(sp1)) == 4))
+		||
+		((!strcmp(lilv_node_as_string(lilv_scale_point_get_label(sp0)), "Cos")
+		  && lilv_node_as_float(lilv_scale_point_get_value(sp0)) == 4)
+		 &&
+		 (!strcmp(lilv_node_as_string(lilv_scale_point_get_label(sp1)), "Sin")
+		  && lilv_node_as_float(lilv_scale_point_get_value(sp1)) == 3)));
 
 	LilvNode* homepage_p = lilv_new_uri(world, "http://usefulinc.com/ns/doap#homepage");
 	LilvNodes* homepages = lilv_plugin_get_value(plug, homepage_p);
@@ -850,7 +969,7 @@ int
 test_ui(void)
 {
 	if (!start_bundle(MANIFEST_PREFIXES
-			":plug a lv2:Plugin ; lv2:binary <foo.so> ; rdfs:seeAlso <plugin.ttl> .\n",
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
 			BUNDLE_PREFIXES PREFIX_LV2UI
 			":plug a lv2:Plugin ; a lv2:CompressorPlugin ; "
 			PLUGIN_NAME("Test plugin") " ; "
@@ -875,11 +994,11 @@ test_ui(void)
 			"] .\n"
 			":ui a lv2ui:GtkUI ; "
 			"  lv2ui:requiredFeature lv2ui:makeResident ; "
-			"  lv2ui:binary <ui.so> ; "
+			"  lv2ui:binary <ui" SHLIB_EXT "> ; "
 			"  lv2ui:optionalFeature lv2ui:ext_presets . "
-			":ui2 a lv2ui:GtkUI ; lv2ui:binary <ui2.so> . "
-			":ui3 a lv2ui:GtkUI ; lv2ui:binary <ui3.so> . "
-			":ui4 a lv2ui:GtkUI ; lv2ui:binary <ui4.so> . "))
+			":ui2 a lv2ui:GtkUI ; lv2ui:binary <ui2" SHLIB_EXT "> . "
+			":ui3 a lv2ui:GtkUI ; lv2ui:binary <ui3" SHLIB_EXT "> . "
+			":ui4 a lv2ui:GtkUI ; lv2ui:binary <ui4" SHLIB_EXT "> . "))
 		return 0;
 
 	init_uris();
@@ -930,7 +1049,7 @@ test_ui(void)
 
 	char* ui_binary_uri_str = (char*)malloc(TEST_PATH_MAX);
 	snprintf(ui_binary_uri_str, TEST_PATH_MAX, "%s%s",
-			lilv_node_as_string(plug_bundle_uri), "ui.so");
+			lilv_node_as_string(plug_bundle_uri), "ui" SHLIB_EXT);
 
 	const LilvNode* ui_binary_uri = lilv_ui_get_binary_uri(ui0);
 
@@ -952,6 +1071,462 @@ test_ui(void)
 
 /*****************************************************************************/
 
+uint32_t atom_Float = 0;
+float    in         = 1.0;
+float    out        = 42.0;
+
+const void*
+get_port_value(const char* port_symbol,
+               void*       user_data,
+               uint32_t*   size,
+               uint32_t*   type)
+{
+	if (!strcmp(port_symbol, "input")) {
+		*size = sizeof(float);
+		*type = atom_Float;
+		return &in;
+	} else if (!strcmp(port_symbol, "output")) {
+		*size = sizeof(float);
+		*type = atom_Float;
+		return &out;
+	} else {
+		fprintf(stderr, "error: get_port_value for nonexistent port `%s'\n",
+		        port_symbol);
+		*size = *type = 0;
+		return NULL;
+	}
+}
+
+void
+set_port_value(const char*     port_symbol,
+               void*           user_data,
+               const void*     value,
+               uint32_t        size,
+               uint32_t        type)
+{
+	if (!strcmp(port_symbol, "input")) {
+		in = *(float*)value;
+	} else if (!strcmp(port_symbol, "output")) {
+		out = *(float*)value;
+	} else {
+		fprintf(stderr, "error: set_port_value for nonexistent port `%s'\n",
+		        port_symbol);
+	}
+}
+
+char** uris   = NULL;
+size_t n_uris = 0;
+
+LV2_URID
+map_uri(LV2_URID_Map_Handle handle,
+        const char*         uri)
+{
+	for (size_t i = 0; i < n_uris; ++i) {
+		if (!strcmp(uris[i], uri)) {
+			return i + 1;
+		}
+	}
+
+	assert(serd_uri_string_has_scheme((const uint8_t*)uri));
+	uris = (char**)realloc(uris, ++n_uris * sizeof(char*));
+	uris[n_uris - 1] = lilv_strdup(uri);
+	return n_uris;
+}
+
+const char*
+unmap_uri(LV2_URID_Map_Handle handle,
+          LV2_URID            urid)
+{
+	if (urid > 0 && urid <= n_uris) {
+		return uris[urid - 1];
+	}
+	return NULL;
+}
+
+static char* temp_dir = NULL;
+
+char*
+lilv_make_path(LV2_State_Make_Path_Handle handle,
+               const char*                path)
+{
+	return lilv_path_join(temp_dir, path);
+}
+
+int
+test_state(void)
+{
+	uint8_t*   abs_bundle = (uint8_t*)lilv_path_absolute(LILV_TEST_BUNDLE);
+	SerdNode   bundle     = serd_node_new_file_uri(abs_bundle, 0, 0, true);
+	LilvWorld* world      = lilv_world_new();
+	LilvNode*  bundle_uri = lilv_new_uri(world, (const char*)bundle.buf);
+	LilvNode*  plugin_uri = lilv_new_uri(world,
+	                                     "http://example.org/lilv-test-plugin");
+	lilv_world_load_bundle(world, bundle_uri);
+	free(abs_bundle);
+	serd_node_free(&bundle);
+
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const LilvPlugin*  plugin  = lilv_plugins_get_by_uri(plugins, plugin_uri);
+	TEST_ASSERT(plugin);
+
+	LV2_URID_Map       map           = { NULL, map_uri };
+	LV2_Feature        map_feature   = { LV2_URID_MAP_URI, &map };
+	LV2_URID_Unmap     unmap         = { NULL, unmap_uri };
+	LV2_Feature        unmap_feature = { LV2_URID_UNMAP_URI, &unmap };
+	const LV2_Feature* features[]    = { &map_feature, &unmap_feature, NULL };
+
+	atom_Float = map.map(map.handle, "http://lv2plug.in/ns/ext/atom#Float");
+
+	LilvNode*  num     = lilv_new_int(world, 5);
+	LilvState* nostate = lilv_state_new_from_file(world, &map, num, "/junk");
+	TEST_ASSERT(!nostate);
+
+	LilvInstance* instance = lilv_plugin_instantiate(plugin, 48000.0, features);
+	TEST_ASSERT(instance);
+	lilv_instance_activate(instance);
+	lilv_instance_connect_port(instance, 0, &in);
+	lilv_instance_connect_port(instance, 1, &out);
+	lilv_instance_run(instance, 1);
+	TEST_ASSERT(in == 1.0);
+	TEST_ASSERT(out == 1.0);
+
+	temp_dir = lilv_realpath("temp");
+
+	const char* file_dir = NULL;
+	char*       copy_dir = NULL;
+	char*       link_dir = NULL;
+	char*       save_dir = NULL;
+
+	// Get instance state state
+	LilvState* state = lilv_state_new_from_instance(
+		plugin, instance, &map,
+		file_dir, copy_dir, link_dir, save_dir,
+		get_port_value, world, 0, NULL);
+
+	// Get another instance state
+	LilvState* state2 = lilv_state_new_from_instance(
+		plugin, instance, &map,
+		file_dir, copy_dir, link_dir, save_dir,
+		get_port_value, world, 0, NULL);
+
+	// Ensure they are equal
+	TEST_ASSERT(lilv_state_equals(state, state2));
+
+	// Save state to a string
+	char* state1_str = lilv_state_to_string(
+		world, &map, &unmap, state, "http://example.org/state1", NULL);
+
+	// Restore from string
+	LilvState* from_str = lilv_state_new_from_string(world, &map, state1_str);
+
+	// Ensure they are equal
+	TEST_ASSERT(lilv_state_equals(state, from_str));
+	free(state1_str);
+
+	const LilvNode* state_plugin_uri = lilv_state_get_plugin_uri(state);
+	TEST_ASSERT(lilv_node_equals(state_plugin_uri, plugin_uri));
+
+	// Tinker with the label of the first state
+	TEST_ASSERT(lilv_state_get_label(state) == NULL);
+	lilv_state_set_label(state, "Test State Old Label");
+	TEST_ASSERT(!strcmp(lilv_state_get_label(state), "Test State Old Label"));
+	lilv_state_set_label(state, "Test State");
+	TEST_ASSERT(!strcmp(lilv_state_get_label(state), "Test State"));
+
+	TEST_ASSERT(!lilv_state_equals(state, state2));  // Label changed
+
+	// Run and get a new instance state (which should now differ)
+	lilv_instance_run(instance, 1);
+	LilvState* state3 = lilv_state_new_from_instance(
+		plugin, instance, &map,
+		file_dir, copy_dir, link_dir, save_dir,
+		get_port_value, world, 0, NULL);
+	TEST_ASSERT(!lilv_state_equals(state2, state3));  // num_runs changed
+
+	// Restore instance state to original state
+	lilv_state_restore(state2, instance, set_port_value, NULL, 0, NULL);
+
+	// Take a new snapshot and ensure it matches the set state
+	LilvState* state4 = lilv_state_new_from_instance(
+		plugin, instance, &map,
+		file_dir, copy_dir, link_dir, save_dir,
+		get_port_value, world, 0, NULL);
+	TEST_ASSERT(lilv_state_equals(state2, state4));
+
+	// Save state to a directory
+	int ret = lilv_state_save(world, &map, &unmap, state, NULL,
+	                          "state.lv2", "state.ttl");
+	TEST_ASSERT(!ret);
+
+	// Load state from directory
+	LilvState* state5 = lilv_state_new_from_file(world, &map, NULL,
+	                                             "state.lv2/state.ttl");
+
+	TEST_ASSERT(lilv_state_equals(state, state5));  // Round trip accuracy
+
+	// Save state with URI to a directory
+	const char* state_uri = "http://example.org/state";
+	ret = lilv_state_save(world, &map, &unmap, state, state_uri,
+	                      "state6.lv2", "state6.ttl");
+	TEST_ASSERT(!ret);
+
+	// Load default bundle into world and load state from it
+	uint8_t*  state6_path       = (uint8_t*)lilv_path_absolute("state6.lv2/");
+	SerdNode  state6_uri        = serd_node_new_file_uri(state6_path, 0, 0, true);
+	LilvNode* test_state_bundle = lilv_new_uri(world, (const char*)state6_uri.buf);
+	LilvNode* test_state_node   = lilv_new_uri(world, state_uri);
+	lilv_world_load_bundle(world, test_state_bundle);
+	lilv_world_load_resource(world, test_state_node);
+	serd_node_free(&state6_uri);
+	free(state6_path);
+
+	LilvState* state6 = lilv_state_new_from_world(world, &map, test_state_node);
+	TEST_ASSERT(lilv_state_equals(state, state6));  // Round trip accuracy
+	lilv_node_free(test_state_bundle);
+	lilv_node_free(test_state_node);
+
+	unsetenv("LV2_STATE_BUNDLE");
+
+	// Make directories and test files support
+	mkdir("temp", 0700);
+	file_dir = temp_dir;
+	mkdir("files", 0700);
+	copy_dir = lilv_realpath("files");
+	mkdir("links", 0700);
+	link_dir = lilv_realpath("links");
+
+	LV2_State_Make_Path make_path         = { NULL, lilv_make_path };
+	LV2_Feature         make_path_feature = { LV2_STATE__makePath, &make_path };
+	const LV2_Feature*  ffeatures[]       = { &make_path_feature, &map_feature, NULL };
+
+	lilv_instance_deactivate(instance);
+	lilv_instance_free(instance);
+	instance = lilv_plugin_instantiate(plugin, 48000.0, ffeatures);
+	lilv_instance_activate(instance);
+	lilv_instance_connect_port(instance, 0, &in);
+	lilv_instance_connect_port(instance, 1, &out);
+	lilv_instance_run(instance, 1);
+
+	// Get instance state state
+	LilvState* fstate = lilv_state_new_from_instance(
+		plugin, instance, &map,
+		file_dir, copy_dir, link_dir, "fstate.lv2",
+		get_port_value, world, 0, ffeatures);
+
+	// Get another instance state
+	LilvState* fstate2 = lilv_state_new_from_instance(
+		plugin, instance, &map,
+		file_dir, copy_dir, link_dir, "fstate2.lv2",
+		get_port_value, world, 0, ffeatures);
+
+	// Should be identical
+	TEST_ASSERT(lilv_state_equals(fstate, fstate2));
+
+	// Run, writing more to rec file
+	lilv_instance_run(instance, 2);
+
+	// Get yet another instance state
+	LilvState* fstate3 = lilv_state_new_from_instance(
+		plugin, instance, &map, file_dir, copy_dir, link_dir, "fstate3.lv2",
+		get_port_value, world, 0, ffeatures);
+
+	// Should be different
+	TEST_ASSERT(!lilv_state_equals(fstate, fstate3));
+
+	// Save state to a directory
+	ret = lilv_state_save(world, &map, &unmap, fstate, NULL,
+	                      "fstate.lv2", "fstate.ttl");
+	TEST_ASSERT(!ret);
+
+	// Load state from directory
+	LilvState* fstate4 = lilv_state_new_from_file(world, &map, NULL,
+	                                              "fstate.lv2/fstate.ttl");
+	TEST_ASSERT(lilv_state_equals(fstate, fstate4));  // Round trip accuracy
+
+	// Restore instance state to loaded state
+	lilv_state_restore(fstate4, instance, set_port_value, NULL, 0, ffeatures);
+
+	// Take a new snapshot and ensure it matches
+	LilvState* fstate5 = lilv_state_new_from_instance(
+		plugin, instance, &map,
+		file_dir, copy_dir, link_dir, "fstate5.lv2",
+		get_port_value, world, 0, ffeatures);
+	TEST_ASSERT(lilv_state_equals(fstate3, fstate5));
+
+	// Save state to a (different) directory again
+	ret = lilv_state_save(world, &map, &unmap, fstate, NULL,
+	                      "fstate6.lv2", "fstate6.ttl");
+	TEST_ASSERT(!ret);
+
+	// Reload it and ensure it's identical to the other loaded version
+	LilvState* fstate6 = lilv_state_new_from_file(world, &map, NULL,
+	                                              "fstate6.lv2/fstate6.ttl");
+	TEST_ASSERT(lilv_state_equals(fstate4, fstate6));
+
+	// Run, changing rec file (without changing size)
+	lilv_instance_run(instance, 3);
+
+	// Take a new snapshot
+	LilvState* fstate7 = lilv_state_new_from_instance(
+		plugin, instance, &map,
+		file_dir, copy_dir, link_dir, "fstate7.lv2",
+		get_port_value, world, 0, ffeatures);
+	TEST_ASSERT(!lilv_state_equals(fstate6, fstate7));
+
+	// Save the changed state to a (different) directory again
+	ret = lilv_state_save(world, &map, &unmap, fstate7, NULL,
+	                      "fstate7.lv2", "fstate7.ttl");
+	TEST_ASSERT(!ret);
+
+	// Reload it and ensure it's changed
+	LilvState* fstate72 = lilv_state_new_from_file(world, &map, NULL,
+	                                              "fstate7.lv2/fstate7.ttl");
+	TEST_ASSERT(lilv_state_equals(fstate72, fstate7));
+	TEST_ASSERT(!lilv_state_equals(fstate6, fstate72));
+
+	lilv_instance_deactivate(instance);
+	lilv_instance_free(instance);
+
+	lilv_node_free(num);
+
+	lilv_state_free(state);
+	lilv_state_free(from_str);
+	lilv_state_free(state2);
+	lilv_state_free(state3);
+	lilv_state_free(state4);
+	lilv_state_free(state5);
+	lilv_state_free(state6);
+	lilv_state_free(fstate);
+	lilv_state_free(fstate2);
+	lilv_state_free(fstate3);
+	lilv_state_free(fstate4);
+	lilv_state_free(fstate5);
+	lilv_state_free(fstate6);
+	lilv_state_free(fstate7);
+	lilv_state_free(fstate72);
+
+	// Free URI map
+	for (size_t i = 0; i < n_uris; ++i) {
+		free(uris[i]);
+	}
+	free(uris);
+	n_uris = 0;
+
+	lilv_node_free(plugin_uri);
+	lilv_node_free(bundle_uri);
+	lilv_world_free(world);
+	free(link_dir);
+	free(copy_dir);
+	free(temp_dir);
+
+	cleanup_uris();
+	return 1;
+}
+
+/*****************************************************************************/
+
+int
+test_bad_port_symbol(void)
+{
+	if (!start_bundle(MANIFEST_PREFIXES
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
+			BUNDLE_PREFIXES PREFIX_LV2EV
+			":plug a lv2:Plugin ; "
+			PLUGIN_NAME("Test plugin") " ; "
+			LICENSE_GPL " ; "
+			"doap:homepage <http://example.org/someplug> ; "
+			"lv2:port [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 0 ; lv2:symbol \"0invalid\" ;"
+			"  lv2:name \"Invalid\" ; "
+			"] ."))
+		return 0;
+
+	init_uris();
+
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const LilvPlugin*  plug    = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+
+	uint32_t n_ports = lilv_plugin_get_num_ports(plug);
+	TEST_ASSERT(n_ports == 0);
+
+	cleanup_uris();
+	return 1;
+}
+
+/*****************************************************************************/
+
+int
+test_bad_port_index(void)
+{
+	if (!start_bundle(MANIFEST_PREFIXES
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
+			BUNDLE_PREFIXES PREFIX_LV2EV
+			":plug a lv2:Plugin ; "
+			PLUGIN_NAME("Test plugin") " ; "
+			LICENSE_GPL " ; "
+			"doap:homepage <http://example.org/someplug> ; "
+			"lv2:port [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index \"notaninteger\" ; lv2:symbol \"invalid\" ;"
+			"  lv2:name \"Invalid\" ; "
+			"] ."))
+		return 0;
+
+	init_uris();
+
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const LilvPlugin*  plug    = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+
+	uint32_t n_ports = lilv_plugin_get_num_ports(plug);
+	TEST_ASSERT(n_ports == 0);
+
+	cleanup_uris();
+	return 1;
+}
+
+/*****************************************************************************/
+
+int
+test_string(void)
+{
+	char* s = NULL;
+
+	TEST_ASSERT(!strcmp((s = lilv_dirname("/foo/bar")), "/foo")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_dirname("/foo/bar/")), "/foo")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_dirname("/foo///bar/")), "/foo")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_dirname("/foo///bar//")), "/foo")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_dirname("foo")), ".")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_dirname("/foo")), "/")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_dirname("/")), "/")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_dirname("//")), "/")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_relative_to("/a/b", "/a/")), "b")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_relative_to("/a", "/b/c/")), "/a")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_relative_to("/a/b/c", "/a/b/d/")), "../c")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_relative_to("/a/b/c", "/a/b/d/e/")), "../../c")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_join("/a", "b")), "/a/b")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_join("/a", "/b")), "/a/b")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_join("/a/", "/b")), "/a/b")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_join("/a/", "b")), "/a/b")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_path_join("/a", NULL)), "/a/")); free(s);
+
+#ifndef _WIN32
+	setenv("LILV_TEST_1", "test", 1);
+	char* home_foo = lilv_strjoin(getenv("HOME"), "/foo", NULL);
+	TEST_ASSERT(!strcmp((s = lilv_expand("$LILV_TEST_1")), "test")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_expand("~")), getenv("HOME"))); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_expand("~foo")), "~foo")); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_expand("~/foo")), home_foo)); free(s);
+	TEST_ASSERT(!strcmp((s = lilv_expand("$NOT_A_VAR")), "$NOT_A_VAR")); free(s);
+	free(home_foo);
+	unsetenv("LILV_TEST_1");
+#endif
+
+	return 1;
+}
+
+/*****************************************************************************/
+
 /* add tests here */
 static struct TestCase tests[] = {
 	TEST_CASE(utils),
@@ -959,10 +1534,16 @@ static struct TestCase tests[] = {
 	TEST_CASE(verify),
 	TEST_CASE(no_verify),
 	TEST_CASE(discovery),
+	TEST_CASE(lv2_path),
 	TEST_CASE(classes),
 	TEST_CASE(plugin),
 	TEST_CASE(port),
 	TEST_CASE(ui),
+	TEST_CASE(bad_port_symbol),
+	TEST_CASE(bad_port_index),
+	TEST_CASE(bad_port_index),
+	TEST_CASE(string),
+	TEST_CASE(state),
 	{ NULL, NULL }
 };
 
