@@ -14,35 +14,33 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#include <locale.h>
+#include <float.h>
 #include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
+#include "lv2/lv2plug.in/ns/ext/port-groups/port-groups.h"
+#include "lv2/lv2plug.in/ns/ext/presets/presets.h"
+#include "lv2/lv2plug.in/ns/ext/event/event.h"
+
 #include "lilv/lilv.h"
 
-#include "lilv-config.h"
+#include "lilv_config.h"
 
-LilvNode* event_class         = NULL;
+#ifdef _MSC_VER
+#    define isnan _isnan
+#endif
+
+LilvNode* applies_to_pred     = NULL;
 LilvNode* control_class       = NULL;
-LilvNode* in_group_pred       = NULL;
-LilvNode* role_pred           = NULL;
-LilvNode* preset_pred         = NULL;
-LilvNode* title_pred          = NULL;
+LilvNode* event_class         = NULL;
+LilvNode* group_pred          = NULL;
+LilvNode* label_pred          = NULL;
+LilvNode* preset_class        = NULL;
+LilvNode* designation_pred    = NULL;
 LilvNode* supports_event_pred = NULL;
 
-void
-print_group(const LilvPlugin* p,
-            const LilvNode*  group,
-            LilvNode*        type,
-            LilvNode*        symbol)
-{
-	printf("\n\tGroup %s:\n", lilv_node_as_string(group));
-	printf("\t\tType: %s\n", lilv_node_as_string(type));
-	printf("\t\tSymbol: %s\n", lilv_node_as_string(symbol));
-}
-
 void
 print_port(const LilvPlugin* p,
            uint32_t          index,
@@ -62,11 +60,11 @@ print_port(const LilvPlugin* p,
 	bool first = true;
 
 	const LilvNodes* classes = lilv_port_get_classes(p, port);
-	printf("\t\tType:       ");
+	printf("\t\tType:        ");
 	LILV_FOREACH(nodes, i, classes) {
 		const LilvNode* value = lilv_nodes_get(classes, i);
 		if (!first) {
-			printf("\n\t\t            ");
+			printf("\n\t\t             ");
 		}
 		printf("%s", lilv_node_as_uri(value));
 		first = false;
@@ -97,42 +95,42 @@ print_port(const LilvPlugin* p,
 	lilv_scale_points_free(points);
 
 	const LilvNode* sym = lilv_port_get_symbol(p, port);
-	printf("\n\t\tSymbol:     %s\n", lilv_node_as_string(sym));
+	printf("\n\t\tSymbol:      %s\n", lilv_node_as_string(sym));
 
 	LilvNode* name = lilv_port_get_name(p, port);
-	printf("\t\tName:       %s\n", lilv_node_as_string(name));
+	printf("\t\tName:        %s\n", lilv_node_as_string(name));
 	lilv_node_free(name);
 
-	LilvNodes* groups = lilv_port_get_value(p, port, in_group_pred);
-	if (lilv_nodes_size(groups) > 0)
-		printf("\t\tGroup:      %s\n",
-		       lilv_node_as_string(
-			       lilv_nodes_get(groups, lilv_nodes_begin(groups))));
+	LilvNodes* groups = lilv_port_get_value(p, port, group_pred);
+	if (lilv_nodes_size(groups) > 0) {
+		printf("\t\tGroup:       %s\n",
+		       lilv_node_as_string(lilv_nodes_get_first(groups)));
+	}
 	lilv_nodes_free(groups);
 
-	LilvNodes* roles = lilv_port_get_value(p, port, role_pred);
-	if (lilv_nodes_size(roles) > 0)
-		printf("\t\tRole:       %s\n",
-		       lilv_node_as_string(
-			       lilv_nodes_get(roles, lilv_nodes_begin(roles))));
-	lilv_nodes_free(roles);
+	LilvNodes* designations = lilv_port_get_value(p, port, designation_pred);
+	if (lilv_nodes_size(designations) > 0) {
+		printf("\t\tDesignation: %s\n",
+		       lilv_node_as_string(lilv_nodes_get_first(designations)));
+	}
+	lilv_nodes_free(designations);
 
 	if (lilv_port_is_a(p, port, control_class)) {
 		if (!isnan(mins[index]))
-			printf("\t\tMinimum:    %f\n", mins[index]);
+			printf("\t\tMinimum:     %f\n", mins[index]);
 		if (!isnan(mins[index]))
-			printf("\t\tMaximum:    %f\n", maxes[index]);
+			printf("\t\tMaximum:     %f\n", maxes[index]);
 		if (!isnan(mins[index]))
-			printf("\t\tDefault:    %f\n", defaults[index]);
+			printf("\t\tDefault:     %f\n", defaults[index]);
 	}
 
 	LilvNodes* properties = lilv_port_get_properties(p, port);
 	if (lilv_nodes_size(properties) > 0)
-		printf("\t\tProperties: ");
+		printf("\t\tProperties:  ");
 	first = true;
 	LILV_FOREACH(nodes, i, properties) {
 		if (!first) {
-			printf("\t\t            ");
+			printf("\t\t             ");
 		}
 		printf("%s\n", lilv_node_as_uri(lilv_nodes_get(properties, i)));
 		first = false;
@@ -198,23 +196,23 @@ print_plugin(LilvWorld*        world,
 
 	LilvUIs* uis = lilv_plugin_get_uis(p);
 	if (lilv_nodes_size(uis) > 0) {
-		printf("\tUI:                ");
+		printf("\tUIs:\n");
 		LILV_FOREACH(uis, i, uis) {
 			const LilvUI* ui = lilv_uis_get(uis, i);
-			printf("%s\n", lilv_node_as_uri(lilv_ui_get_uri(ui)));
+			printf("\t\t%s\n", lilv_node_as_uri(lilv_ui_get_uri(ui)));
 
 			const char* binary = lilv_node_as_uri(lilv_ui_get_binary_uri(ui));
 
 			const LilvNodes* types = lilv_ui_get_classes(ui);
 			LILV_FOREACH(nodes, i, types) {
-				printf("\t                       Class:  %s\n",
+				printf("\t\t\tClass:  %s\n",
 				       lilv_node_as_uri(lilv_nodes_get(types, i)));
 			}
 
 			if (binary)
-				printf("\t                       Binary: %s\n", binary);
+				printf("\t\t\tBinary: %s\n", binary);
 
-			printf("\t                       Bundle: %s\n",
+			printf("\t\t\tBundle: %s\n",
 			       lilv_node_as_uri(lilv_ui_get_bundle_uri(ui)));
 		}
 	}
@@ -266,28 +264,50 @@ print_plugin(LilvWorld*        world,
 		printf("\n");
 	lilv_nodes_free(features);
 
+	/* Extension Data */
+
+	LilvNodes* data = lilv_plugin_get_extension_data(p);
+	if (data)
+		printf("\tExtension Data:    ");
+	first = true;
+	LILV_FOREACH(nodes, i, data) {
+		if (!first) {
+			printf("\n\t                   ");
+		}
+		printf("%s", lilv_node_as_uri(lilv_nodes_get(data, i)));
+		first = false;
+	}
+	if (data)
+		printf("\n");
+	lilv_nodes_free(data);
+
 	/* Presets */
 
-	LilvNodes* presets = lilv_plugin_get_value(p, preset_pred);
+	LilvNodes* presets = lilv_plugin_get_related(p, preset_class);
 	if (presets)
 		printf("\tPresets: \n");
 	LILV_FOREACH(nodes, i, presets) {
-		LilvNodes* titles = lilv_world_find_nodes(world,
-		                                          lilv_nodes_get(presets, i),
-		                                          title_pred,
-		                                          NULL);
+		const LilvNode* preset = lilv_nodes_get(presets, i);
+		lilv_world_load_resource(world, preset);
+		LilvNodes* titles = lilv_world_find_nodes(
+			world, preset, label_pred, NULL);
 		if (titles) {
-			const LilvNode* title = lilv_nodes_get(titles, lilv_nodes_begin(titles));
+			const LilvNode* title = lilv_nodes_get_first(titles);
 			printf("\t         %s\n", lilv_node_as_string(title));
+			lilv_nodes_free(titles);
+		} else {
+			fprintf(stderr, "Preset <%s> has no rdfs:label\n",
+			        lilv_node_as_string(lilv_nodes_get(presets, i)));
 		}
 	}
+	lilv_nodes_free(presets);
 
 	/* Ports */
 
 	const uint32_t num_ports = lilv_plugin_get_num_ports(p);
-	float* mins     = calloc(num_ports, sizeof(float));
-	float* maxes    = calloc(num_ports, sizeof(float));
-	float* defaults = calloc(num_ports, sizeof(float));
+	float* mins     = (float*)calloc(num_ports, sizeof(float));
+	float* maxes    = (float*)calloc(num_ports, sizeof(float));
+	float* defaults = (float*)calloc(num_ports, sizeof(float));
 	lilv_plugin_get_port_ranges_float(p, mins, maxes, defaults);
 
 	for (uint32_t i = 0; i < num_ports; ++i)
@@ -313,12 +333,12 @@ void
 print_usage(void)
 {
 	printf(
-		"Usage: lv2info [OPTIONS] PLUGIN_URI\n"
-		"Show information about an installed LV2 plugin.\n\n"
+		"Usage: lv2info [OPTION]... PLUGIN_URI\n"
+		"Print information about an installed LV2 plugin.\n\n"
 		"  -p FILE      Write Turtle description of plugin to FILE\n"
 		"  -m FILE      Add record of plugin to manifest FILE\n"
 		"  --help       Display this help and exit\n"
-		"  --version    Output version information and exit\n\n"
+		"  --version    Display version information and exit\n\n"
 		"For -p and -m, Turtle files are appended to (not overwritten),\n"
 		"and @prefix directives are only written if the file was empty.\n"
 		"This allows several plugins to be added to a single file.\n");
@@ -355,23 +375,18 @@ main(int argc, char** argv)
 	}
 
 	int ret = 0;
-	setlocale(LC_ALL, "");
 
 	LilvWorld* world = lilv_world_new();
 	lilv_world_load_all(world);
 
-#define NS_DC   "http://dublincore.org/documents/dcmi-namespace/"
-#define NS_PG   "http://lv2plug.in/ns/ext/port-groups#"
-#define NS_PSET "http://lv2plug.in/ns/ext/presets#"
-#define NS_EV   "http://lv2plug.in/ns/ext/event#"
-
+	applies_to_pred     = lilv_new_uri(world, LV2_CORE__appliesTo);
 	control_class       = lilv_new_uri(world, LILV_URI_CONTROL_PORT);
 	event_class         = lilv_new_uri(world, LILV_URI_EVENT_PORT);
-	in_group_pred       = lilv_new_uri(world, NS_PG "inGroup");
-	preset_pred         = lilv_new_uri(world, NS_PSET "hasPreset");
-	role_pred           = lilv_new_uri(world, NS_PG "role");
-	title_pred          = lilv_new_uri(world, NS_DC "title");
-	supports_event_pred = lilv_new_uri(world, NS_EV "supportsEvent");
+	group_pred          = lilv_new_uri(world, LV2_PORT_GROUPS__group);
+	label_pred          = lilv_new_uri(world, LILV_NS_RDFS "label");
+	preset_class        = lilv_new_uri(world, LV2_PRESETS__Preset);
+	designation_pred    = lilv_new_uri(world, LV2_CORE__designation);
+	supports_event_pred = lilv_new_uri(world, LV2_EVENT__supportsEvent);
 
 	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
 	LilvNode*          uri     = lilv_new_uri(world, plugin_uri);
@@ -401,12 +416,15 @@ main(int argc, char** argv)
 
 	lilv_node_free(uri);
 
-	lilv_node_free(title_pred);
-	lilv_node_free(role_pred);
-	lilv_node_free(preset_pred);
-	lilv_node_free(in_group_pred);
+	lilv_node_free(supports_event_pred);
+	lilv_node_free(designation_pred);
+	lilv_node_free(preset_class);
+	lilv_node_free(label_pred);
+	lilv_node_free(group_pred);
 	lilv_node_free(event_class);
 	lilv_node_free(control_class);
+	lilv_node_free(applies_to_pred);
+
 	lilv_world_free(world);
 	return ret;
 }
